"""
Example code that just spits out random numbers between 0 and 3
for z_mode, and Gaussian centered at z_mode with width
random_width*(1+zmode).
"""

from typing import Any

import numpy as np
import qp
from ceci.config import StageParameter as Param
from scipy.stats import norm

from rail.core.data import ModelHandle, TableHandle, TableLike
from rail.estimation.estimator import CatEstimator
from rail.estimation.informer import CatInformer


class RandomGaussInformer(CatInformer):
    """Placeholder Informer"""

    name = "RandomGaussInformer"
    config_options = CatInformer.config_options.copy()

    def run(self) -> None:
        self.add_data("model", np.array([None]))


class RandomGaussEstimator(CatEstimator):
    """Random CatEstimator"""

    name = "RandomGaussEstimator"
    inputs = [("input", TableHandle), ("model", ModelHandle)]
    config_options = CatEstimator.config_options.copy()
    config_options.update(
        rand_width=Param(float, 0.025, "ad hock width of PDF"),
        rand_zmin=Param(float, 0.0, msg="The minimum redshift of the z grid"),
        rand_zmax=Param(float, 3.0, msg="The maximum redshift of the z grid"),
        nzbins=Param(int, 301, msg="The number of gridpoints in the z grid"),
        seed=Param(int, 87, msg="random seed"),
        column_name=Param(
            str,
            "mag_i_lsst",
            msg="name of a column that has the "
            "correct number of galaxies to find length of",
        ),
    )

    def __init__(self, args: Any, **kwargs: Any) -> None:
        """Constructor:
        Do CatEstimator specific initialization"""
        super().__init__(args, **kwargs)
        self.zgrid = None

    def _process_chunk(
        self, start: int, end: int, data: TableLike, first: bool
    ) -> None:
        pdf = []
        # allow for either format for now
        numzs = len(data[self.config.column_name])
        rng = np.random.default_rng(seed=self.config.seed + start)
        zmode = np.round(rng.uniform(0.0, self.config.rand_zmax, numzs), 3)
        widths = self.config.rand_width * (1.0 + zmode)
        self.zgrid = np.linspace(
            self.config.rand_zmin, self.config.rand_zmax, self.config.nzbins
        )
        for i in range(numzs):
            pdf.append(norm.pdf(self.zgrid, zmode[i], widths[i]))
        qp_d = qp.Ensemble(
            qp.stats.norm,  # pylint: disable=no-member
            data=dict(
                loc=np.expand_dims(zmode, -1),  # pylint: disable=no-member
                scale=np.expand_dims(widths, -1),
            ),
        )
        qp_d.set_ancil(dict(zmode=zmode))
        self._do_chunk_output(qp_d, start, end, first, data=data)

    def validate(self) -> None:
        """Validation which checks if the required column names by the stage exist in the data"""
        self._get_stage_columns()
        data = self.get_handle("input", allow_missing=True)
        # **kwargs in the function below is omitted
        # as these params are not yet implemented
        self._check_column_names(data, self.stage_columns)

    def _get_stage_columns(self) -> None:
        self.stage_columns = [self.config.column_name]



"""Add a column of random numbers to a dataframe."""

from typing import Any

import numpy as np
from ceci.config import StageParameter as Param

from rail.creation.noisifier import Noisifier


class AddColumnOfRandom(Noisifier):
    """Add a column of random numbers to a dataframe"""

    name = "AddColumnOfRandom"
    config_options = Noisifier.config_options.copy()
    config_options.update(
        col_name=Param(
            str, "chaos_bunny", msg="Name of the column with random numbers"
        ),
    )

    def __init__(self, args: Any, **kwargs: Any) -> None:
        """
        Constructor

        Does standard Noisifier initialization
        """
        Noisifier.__init__(self, args, **kwargs)

    def _initNoiseModel(self) -> None:  # pragma: no cover
        np.random.seed(self.config.seed)

    def _addNoise(self) -> None:  # pragma: no cover
        data = self.get_data("input")
        copy = data.copy()
        copy.insert(0, self.config.col_name, np.random.uniform(size=len(copy)))
        self.add_data("output", copy)



"""Wrapper class for ML Models"""

from __future__ import annotations

import pickle
from typing import Any


class Model:
    """Class to act as wrapper for ML models

    This will attach metadata to the ML model, and provide tools
    for validation, versioning and tracking provenance.

    """

    def __init__(
        self,
        data: Any,
        creation_class_name: str,
        version: int = 0,
        catalog_tag: str | None = None,
        provenance: dict | None = None,
    ) -> None:
        """Constructor

        Parameters
        ----------
        data
            Model data

        creation_class_name
            Name of class that created this model

        version
            Version of the model

        catalog_tag
            Associated CatalogConfigBase that defined training dataset

        provenance
            Provenance infomration
        """
        self.data = data
        self.creation_class_name = creation_class_name
        self.version = version
        self.catalog_tag = catalog_tag
        if provenance is not None:
            self.provenance = provenance.copy()
        else:
            self.provenance = {}

    def validate(self, creation_class_name: str | None, version: int | None) -> None:
        """
        Parameters
        ----------
        creation_class_name
            Name of class that created this model

        version
            Version of the model

        Raises
        ------
        TypeError : Either creation_class_name or version does not match
        """
        if (
            creation_class_name is not None
            and creation_class_name != self.creation_class_name
        ):
            raise TypeError(
                "Model.creation_class_name does not match.  "
                f"{creation_class_name} != {self.creation_class_name}"
            )
        if version is not None and version != self.version:
            raise TypeError(
                f"Model.version does not match.  {version} != {self.version}"
            )

    @classmethod
    def read(
        cls,
        path: str,
        creation_class_name: str = "dummy",
        version: int = 0,
        catalog_tag: str | None = None,
        provenance: dict | None = None,
    ) -> Model:
        """Read a model from a file.

        Note that this will promote the data to a Model if it is not already

        Parameters
        ----------
        path
            File to read

        creation_class_name:
            Name of class that created this model

        version:
            Version of the model

        catalog_tag
            Associated CatalogConfigBase that defined training dataset

        provenance:
            Provenance infomration

        Returns
        -------
        Model
            Newly read Model
        """
        with open(path, "rb") as fin:
            read_data = pickle.load(fin)

        if isinstance(read_data, Model):
            return read_data

        if provenance is None:
            provenance = {}

        return cls(read_data, creation_class_name, version, catalog_tag, provenance)

    @classmethod
    def wrap(
        cls,
        inpath: str,
        outpath: str,
        creation_class_name: str = "dummy",
        version: int = 0,
        catalog_tag: str | None = None,
        provenance: dict | None = None,
    ) -> Model:
        """Read a model from a file and write it as a `Model` if it is not already

        Parameters
        ----------
        inpath
            File to read

        outpath:
            File to write

        creation_class_name:
            Name of class that created this model

        version:
            Version of the model

        catalog_tag
            Associated CatalogConfigBase that defined training dataset

        provenance:
            Provenance information

        Returns
        -------
        Model
            Newly read & converted Model
        """
        the_model = cls.read(
            inpath, creation_class_name, version, catalog_tag, provenance
        )
        the_model.write(outpath)
        return the_model

    @classmethod
    def dump(
        cls,
        obj: Any,
        path: str,
        creation_class_name: str = "dummy",
        version: int = 0,
        catalog_tag: str | None = None,
        provenance: dict | None = None,
    ) -> Model:
        """Write an object to a model file

        This will promote it to a Model if it isn't already

        Parameters
        ----------
        obj
            Object to dump

        path:
            File to write

        creation_class_name:
            Name of class that created this model

        version:
            Version of the model

        catalog_tag
            Associated CatalogConfigBase that defined training dataset

        provenance:
            Provenance information

        Returns
        -------
        Model
            Newly converted Model
        """

        if isinstance(obj, Model):
            write_obj = obj
        else:
            if provenance is None:
                provenance = {}
            write_obj = cls(obj, creation_class_name, version, catalog_tag, provenance)

        write_obj.write(path)
        return write_obj

    def write(
        self,
        path: str,
    ) -> None:
        """Write a model to a file

        Parameters
        ----------
        path
            File to write
        """
        with open(path, "wb") as fout:
            pickle.dump(obj=self, file=fout, protocol=pickle.HIGHEST_PROTOCOL)



"""
Implementation of the 'pathological photo-z PDF estimator,
as used in arXiv:2001.03621 (see section 3.3). It assigns each test set galaxy
a photo-z PDF equal to the normalized redshift distribution
N (z) of the training set.
"""

from typing import Any

import numpy as np
import qp

from rail.core.common_params import SHARED_PARAMS
from rail.core.data import TableLike
from rail.estimation.estimator import CatEstimator
from rail.estimation.informer import CatInformer


class trainZmodel:
    """
    Temporary class to store the single trainZ pdf for trained model.
    Given how simple this is to compute, this seems like overkill.
    """

    def __init__(self, zgrid: np.ndarray, pdf: np.ndarray, zmode: np.ndarray) -> None:
        self.zgrid = zgrid
        self.pdf = pdf
        self.zmode = zmode


class TrainZInformer(CatInformer):
    """Train an Estimator which returns a global PDF for all galaxies"""

    name = "TrainZInformer"
    config_options = CatInformer.config_options.copy()
    config_options.update(
        zmin=SHARED_PARAMS,
        zmax=SHARED_PARAMS,
        nzbins=SHARED_PARAMS,
        redshift_col=SHARED_PARAMS,
    )

    def run(self) -> None:
        if self.config.hdf5_groupname:
            training_data = self.get_data("input")[self.config.hdf5_groupname]
        else:  # pragma: no cover
            training_data = self.get_data("input")
        zbins = np.linspace(self.config.zmin, self.config.zmax, self.config.nzbins + 1)
        speczs = np.sort(training_data[self.config.redshift_col])
        train_pdf, _ = np.histogram(speczs, zbins)
        midpoints = zbins[:-1] + np.diff(zbins) / 2
        zmode = midpoints[np.argmax(train_pdf)]
        cdf = np.cumsum(train_pdf)
        cdf = cdf / cdf[-1]
        norm = cdf[-1] * (zbins[2] - zbins[1])
        train_pdf = train_pdf / norm
        zgrid = midpoints
        self.model = trainZmodel(zgrid, train_pdf, zmode)
        self.add_data("model", self.model)

    def validate(self) -> None:
        """Validation which checks if the required column names by the stage exist in the data"""
        self._get_stage_columns()
        data = self.get_handle("input", allow_missing=True)
        # **kwargs in the function below is omitted
        # as these params are not yet implemented
        self._check_column_names(data, self.stage_columns)

    def _get_stage_columns(self) -> None:
        self.stage_columns = [self.config.redshift_col]


class TrainZEstimator(CatEstimator):
    """CatEstimator which returns a global PDF for all galaxies"""

    name = "TrainZEstimator"
    config_options = CatEstimator.config_options.copy()
    config_options.update(zmin=SHARED_PARAMS, zmax=SHARED_PARAMS, nzbins=SHARED_PARAMS)

    def __init__(self, args: Any, **kwargs: Any) -> None:
        self.zgrid: np.ndarray | None = None
        self.train_pdf: np.ndarray | None = None
        self.zmode: np.ndarray | None = None
        CatEstimator.__init__(self, args, **kwargs)

    def open_model(self, **kwargs: Any) -> None:
        CatEstimator.open_model(self, **kwargs)
        if self.model is None:  # pragma: no cover
            return
        self.zgrid = self.model.zgrid
        self.train_pdf = self.model.pdf
        self.zmode = self.model.zmode

    def _process_chunk(
        self, start: int, end: int, data: TableLike, first: bool
    ) -> None:
        test_size = end - start
        assert self.zmode is not None
        assert self.train_pdf is not None
        zmode = np.repeat(self.zmode, test_size)
        qp_d = qp.Ensemble(
            qp.interp,
            data=dict(xvals=self.zgrid, yvals=np.tile(self.train_pdf, (test_size, 1))),
        )
        qp_d.set_ancil(dict(zmode=zmode))
        self._do_chunk_output(qp_d, start, end, first, data=data)



import os

import numpy as np
import pytest

from rail.core.data import QPHandle
from rail.core.stage import RailStage
from rail.estimation.algos.equal_count import EqualCountClassifier
from rail.estimation.algos.uniform_binning import UniformBinningClassifier
from rail.utils.path_utils import RAILDIR

DS = RailStage.data_store
DS.__class__.allow_overwrite = True

inputdata = os.path.join(RAILDIR, "rail/examples_data/testdata/output_BPZ_lite.hdf5")


@pytest.mark.parametrize(
    "input_param",
    [
        {"zbin_edges": [0.0, 0.3]},
        {"zmin": 0.0, "zmax": 0.3, "nbins": 1},
        {"zbin_edges": [0.0, 0.3], "id_name": "CATAID"},
    ],
)
def test_UniformBinningClassifier(input_param: dict) -> None:
    DS.clear()
    input_data = DS.read_file("input_data", QPHandle, inputdata)

    tomo = UniformBinningClassifier.make_stage(
        point_estimate="zmode",
        no_assign=-99,
        **input_param,
    )

    _out_data = tomo.classify(input_data)
    os.remove(tomo.get_output(tomo.get_aliased_tag("output"), final_name=True))


def test_UniformBinningClassifier_binsize() -> None:
    DS.clear()
    input_data = DS.read_file("input_data", QPHandle, inputdata)

    tomo = UniformBinningClassifier.make_stage(
        point_estimate="zmode",
        no_assign=-99,
        zmin=0.0,
        zmax=2.0,
        nbins=2,
    )
    output_data = tomo.classify(input_data)
    out_data = output_data.data

    # check length:
    assert len(out_data["class_id"]) == len(out_data["row_index"])

    # check that the assignment is as expected:
    assert (np.in1d(np.unique(out_data["class_id"]), [1, 2, -99])).all()

    zb = input_data.data.ancil["zmode"]
    if 1 in out_data["class_id"]:
        assert (
            (zb[out_data["class_id"] == 1] >= 0.0)
            & (zb[out_data["class_id"] == 1] < 1.0)
        ).all()
    if 2 in out_data["class_id"]:
        assert (
            (zb[out_data["class_id"] == 2] >= 1.0)
            & (zb[out_data["class_id"] == 2] < 2.0)
        ).all()
    if -99 in out_data["class_id"]:
        assert (
            (zb[out_data["class_id"] == -99] < 0.0)
            | (zb[out_data["class_id"] == -99] >= 2.0)
        ).all()
    os.remove(tomo.get_output(tomo.get_aliased_tag("output"), final_name=True))


def test_UniformBinningClassifier_ancil() -> None:
    DS.clear()
    input_data = DS.read_file("input_data", QPHandle, inputdata)

    tomo = UniformBinningClassifier.make_stage(
        point_estimate="zmedian",
        no_assign=-99,
        zmin=0.0,
        zmax=2.0,
        nbins=2,
    )
    with pytest.raises(KeyError):
        _out_data = tomo.classify(input_data)


@pytest.mark.parametrize(
    "input_param",
    [
        {"zmin": 0.0, "zmax": 0.3, "nbins": 1},
        {"zmin": 0.0, "zmax": 0.3, "nbins": 1, "id_name": "CATAID"},
    ],
)
def test_EqualCountClassifier(input_param: dict) -> None:
    DS.clear()
    input_data = DS.read_file("input_data", QPHandle, inputdata)

    tomo = EqualCountClassifier.make_stage(
        point_estimate="zmode",
        no_assign=-99,
        **input_param,
    )

    _out_data = tomo.classify(input_data)
    os.remove(tomo.get_output(tomo.get_aliased_tag("output"), final_name=True))


def test_EqualCountClassifier_nobj() -> None:
    DS.clear()
    input_data = DS.read_file("input_data", QPHandle, inputdata)

    tomo = EqualCountClassifier.make_stage(
        point_estimate="zmode",
        no_assign=-99,
        zmin=0.0,
        zmax=2.0,
        nbins=2,
    )
    output_data = tomo.classify(input_data)
    out_data = output_data.data

    # check that there are equal number of object in each bin modulo Ngal%Nbins
    assert (np.in1d(np.unique(out_data["class_id"]), [1, 2, -99])).all()

    Ngal = sum(out_data["class_id"] != -99)
    exp_Ngal_perbin = int(Ngal / 2)
    # check that each bin does contain number of objects consistent with expected number
    # exp_Ngal_perbin + 1 to account for the cases where Ngal%Nbins!=0
    assert sum(out_data["class_id"] == 1) in [exp_Ngal_perbin, exp_Ngal_perbin + 1]
    assert sum(out_data["class_id"] == 2) in [exp_Ngal_perbin, exp_Ngal_perbin + 1]

    # check no assignment is correct
    if Ngal < len(out_data["class_id"]):
        zb = input_data.data.ancil["zmode"]
        assert (
            (zb[out_data["class_id"] == -99] < 0.0)
            | (zb[out_data["class_id"] == -99] >= 2.0)
        ).all()
    os.remove(tomo.get_output(tomo.get_aliased_tag("output"), final_name=True))


def test_EqualCountClassifier_ancil() -> None:
    DS.clear()
    input_data = DS.read_file("input_data", QPHandle, inputdata)

    tomo = EqualCountClassifier.make_stage(
        point_estimate="zmedian",
        no_assign=-99,
        zmin=0.0,
        zmax=2.0,
        nbins=2,
    )
    with pytest.raises(KeyError):
        _out_data = tomo.classify(input_data)



import os

import numpy as np
import pytest

from rail.interfaces import PZFactory
from rail.utils.path_utils import find_rail_file


def test_pz_factory() -> None:
    stage = PZFactory.build_cat_estimator_stage(
        "train_z",
        "TrainZEstimator",
        "rail.estimation.algos.train_z",
        "tests/interfaces/model_inform_trainz.pkl",
        "dummy.in",
    )

    input_file = find_rail_file("examples_data/testdata/validation_10gal.hdf5")

    out_single = PZFactory.estimate_single_pz(stage, {"d": np.array([1, 1])})
    assert out_single.npdf == 1

    _out_handle = PZFactory.run_cat_estimator_stage(
        stage,
        input_file,
    )

    check_stage = PZFactory.get_cat_estimator_stage("train_z")
    assert check_stage == stage

    with pytest.raises(KeyError):
        PZFactory.get_cat_estimator_stage("nope")

    PZFactory.reset()
    assert not PZFactory._stage_dict

    try:
        os.unlink("inprogress_output_train_z.hdf5")
    except FileNotFoundError:
        pass
    try:
        os.unlink("output_train_z.hdf5")
    except FileNotFoundError:
        pass



from typing import Any

import numpy as np
from ceci.config import StageParameter as Param
from qp.metrics.base_metric_classes import MetricOutputType
from qp.metrics.point_estimate_metric_classes import PointToPointMetric

from rail.core.data import TableLike, QPHandle, TableHandle
from rail.evaluation.evaluator import Evaluator


class PointToPointEvaluator(Evaluator):
    """Evaluate the performance of a photo-z estimator against reference point estimate"""

    name = "PointToPointEvaluator"
    config_options = Evaluator.config_options.copy()
    config_options.update(
        hdf5_groupname=Param(
            str, "photometry", required=False, msg="HDF5 Groupname for truth table."
        ),
        reference_dictionary_key=Param(
            str,
            "redshift",
            required=False,
            msg="The key in the `truth` dictionary where the redshift data is stored.",
        ),
        point_estimate_key=Param(
            str, "zmode", required=False, msg="The key in the point estimate table."
        ),
    )
    inputs = [("input", QPHandle), ("truth", TableHandle)]

    metric_base_class = PointToPointMetric

    def _process_chunk(self, data_tuple: Any, first: bool) -> None:
        start = data_tuple[0]
        end = data_tuple[1]
        estimate_data = np.squeeze(data_tuple[2].ancil[self.config.point_estimate_key])
        reference_data = data_tuple[3][self.config.reference_dictionary_key]

        self._process_all_chunk_metrics(
            estimate_data, reference_data, start, end, first
        )

    def _process_all(self, data_tuple: Any) -> None:
        estimate_data = np.squeeze(data_tuple[0].ancil[self.config.point_estimate_key])
        reference_data = data_tuple[1][self.config.hdf5_groupname][
            self.config.reference_dictionary_key
        ]

        self._process_all_metrics(estimate_data, reference_data)


class PointToPointBinnedEvaluator(Evaluator):
    """Evaluate the performance of a photo-z estimator against reference point estimate"""

    name = "PointToPointBinnedEvaluator"
    config_options = Evaluator.config_options.copy()
    config_options.update(
        hdf5_groupname=Param(
            str, "photometry", required=False, msg="HDF5 Groupname for truth table."
        ),
        reference_dictionary_key=Param(
            str,
            "redshift",
            required=False,
            msg="The key in the `truth` dictionary where the redshift data is stored.",
        ),
        point_estimate_key=Param(
            str, "zmode", required=False, msg="The key in the point estimate table."
        ),
        bin_col=Param(
            str, "redshift", required=False, msg="The column metrics are binned by"
        ),
        bin_min=Param(
            float, 0.0, required=False, msg="The mininum value of the binning edge"
        ),
        bin_max=Param(
            float, 3.0, required=False, msg="The maximum value of the binning edge"
        ),
        nbin=Param(
            int, 10, required=False, msg="The mininum value of the binning edge"
        ),
        force_exact=Param(
            bool,
            default=True,
            required=False,
            msg="Force the exact calculation.  This will not allow parallelization",
        ),
    )
    inputs = [("input", QPHandle), ("truth", TableHandle)]

    metric_base_class = PointToPointMetric

    def run(self) -> None:
        self._build_config_dict()

        print(f"Requested metrics: {list(self._metric_config_dict.keys())}")
        data = self._get_all_data()
        estimate_data = np.squeeze(data[0].ancil[self.config.point_estimate_key])
        reference_data = data[1][self.config.hdf5_groupname][
            self.config.reference_dictionary_key
        ]
        bin_edges = np.linspace(
            self.config.bin_min, self.config.bin_max, self.config.nbin
        )
        bin_col = data[1][self.config.hdf5_groupname][self.config.bin_col]
        bin_indices = np.digitize(bin_col, bins=bin_edges) - 1
        self.run_single_node()
        table_list = []

        for i in range(len(bin_edges) - 1):
            estimate_data_subset = estimate_data[bin_indices == i]
            reference_data_subset = reference_data[bin_indices == i]

            summary_table = self._process_all_metrics_binned(
                estimate_data_subset, reference_data_subset
            )
            table_list.append(summary_table)

        out_dict = {}
        out_dict["bin_center"] = (bin_edges[:-1] + bin_edges[1:]) / 2
        metric_keys = table_list[0].keys()
        for key in metric_keys:
            out_dict[key] = np.array(
                [table_list[j][key][0] for j in range(len(table_list))]
            )
        self._summary_handle = self.add_handle("summary", data=out_dict)

    def _process_all_metrics_binned(
        self, estimate_data: Any, reference_data: Any
    ) -> TableLike:
        """This function writes out metric values when operating in non-parallel mode.

        Parameters
        ----------
        estimate_data
            The estimated values (of the appropriate type, float or pdf) to be used
            by the requested metrics.

        reference_data
            The reference or known values (of the appropriate type, float or pdf)
            to be used by the requested metrics.

        Raises
        ------
        ValueError
            Raises an error if an unknown metric is requested.
        """

        summary_table = {}

        for metric, this_metric in self._cached_metrics.items():
            if metric not in self._metric_dict:  # pragma: no cover
                raise ValueError(
                    f"Unsupported metric requested: '{metric}'. "
                    f"Available metrics are: {sorted(self._metric_dict.keys())}"
                )

            metric_result = this_metric.evaluate(estimate_data, reference_data)

            if this_metric.metric_output_type == MetricOutputType.single_value:
                summary_table[metric] = np.array([metric_result])

        return summary_table

    def _process_all(self, data_tuple: Any) -> None:
        estimate_data = np.squeeze(data_tuple[0].ancil[self.config.point_estimate_key])
        reference_data = data_tuple[1][self.config.hdf5_groupname][
            self.config.reference_dictionary_key
        ]

        self._process_all_metrics(estimate_data, reference_data)



import enum
from functools import partial
from typing import Any, Type, TypeVar

import click

__all__ = [
    "args_to_dict",
    "clear_output",
    "bpz_demo_data",
    "catalog_tag",
    "dry_run",
    "outdir",
    "output_yaml",
    "from_source",
    "model_file",
    "git_mode",
    "params",
    "pipeline_class",
    "pipeline_yaml",
    "package_file",
    "skip",
    "stage_class",
    "stage_module",
    "stage_name",
    "stages_config",
    "inputs",
    "verbose_download",
]


class GitMode(enum.Enum):
    """Choose git clone mode"""

    ssh = 0
    https = 1
    cli = 2


EnumType_co = TypeVar("EnumType_co", bound=Type[enum.Enum], covariant=True)


class EnumChoice(click.Choice):
    """A version of click.Choice specialized for enum types"""

    def __init__(self, the_enum: EnumType_co, case_sensitive: bool = True) -> None:
        self._enum = the_enum
        super().__init__(
            list(the_enum.__members__.keys()), case_sensitive=case_sensitive
        )

    def convert(
        self, value: Any, param: click.Parameter | None, ctx: click.Context | None
    ) -> enum.Enum:  # pragma: no cover
        converted_str = super().convert(value, param, ctx)
        return self._enum.__members__[converted_str]


def args_to_dict(args: list[str]) -> dict[str, str]:
    """Convert a series of command line key=value statements
    to a dict"""
    out_dict = {}
    for arg_ in args:
        tokens = arg_.split("=")
        if len(tokens) != 2:
            raise ValueError(f"Poorly formed argument {arg_}.  Should by key=value")
        out_dict[tokens[0]] = tokens[1]
    return out_dict


class PartialOption:
    """Wraps click.option with partial arguments for convenient reuse"""

    def __init__(self, *param_decls: str, **kwargs: Any) -> None:
        self._partial = partial(click.option, *param_decls, cls=click.Option, **kwargs)

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        return self._partial(*args, **kwargs)


class PartialArgument:
    """Wraps click.argument with partial arguments for convenient reuse"""

    def __init__(self, *param_decls: Any, **kwargs: Any) -> None:
        self._partial = partial(
            click.argument, *param_decls, cls=click.Argument, **kwargs
        )

    def __call__(self, *args: Any, **kwargs: Any) -> Any:  # pragma: no cover
        return self._partial(*args, **kwargs)


clear_output = PartialOption(
    "--clear-output",
    help="Clear Notebook output",
    is_flag=True,
)

dry_run = PartialOption(
    "--dry-run",
    help="Dry run only",
    is_flag=True,
)


catalog_tag = PartialOption(
    "--catalog_tag",
    default=None,
    help="Type of input catalog, used to determine column names",
)


from_source = PartialOption(
    "--from-source",
    help="Install from source",
    is_flag=True,
)

outdir = PartialOption(
    "--outdir",
    type=click.Path(),
    default=None,
    help="Output directory.",
)

output_yaml = PartialOption(
    "--output_yaml",
    type=click.Path(),
    default=None,
    help="Path for output yaml file",
)

pipeline_class = PartialOption(
    "--pipeline_class",
    type=str,
    help="Full class name for pipeline, e.g., rail.pipelines.estimation.train_z.TrainZPipeline",
)

model_file = PartialOption(
    "--model_file",
    type=str,
    help="Model for pz estimation",
)

input_file = PartialOption(
    "--input_file",
    type=str,
    help="Input data file for pz estimation",
)

pipeline_yaml = PartialOption(
    "--pipeline_yaml",
    type=click.Path(),
    help="Yaml for that defines pipeline",
)

git_mode = PartialOption(
    "--git-mode",
    type=EnumChoice(GitMode),
    default="ssh",
    help="Git Clone mode",
)

project_yaml = PartialOption(
    "--project_yaml",
    type=click.Path(),
    default=None,
    help="File with project description",
)

package_file = PartialOption(
    "--package-file",
    type=click.Path(),
    default=None,
    help="File with package",
)

skip = PartialOption(
    "--skip",
    type=click.Path(),
    multiple=True,
    help="Skip files",
)

stage_class = PartialOption(
    "--stage_class",
    type=str,
    help="Name of a pipeline stage python class",
)

stage_module = PartialOption(
    "--stage_module",
    type=str,
    help="Import path for a python module",
)

stage_name = PartialOption(
    "--stage_name",
    type=str,
    help="Name of a pipeline stage",
)

stages_config = PartialOption(
    "--stages_config",
    type=str,
    help="Stage config file",
    default=None,
)

inputs = PartialArgument("inputs", nargs=-1)

params = PartialArgument("params", nargs=-1)

verbose_download = PartialOption(
    "-v", "--verbose", help="Verbose output when downloading", is_flag=True
)

bpz_demo_data = PartialOption(
    "--bpz-demo-data",
    help="Download data that is explicitly only for use in the bpz demo and nowhere else"
    "(it is dummy data that will not make sense)",
    is_flag=True,
)



from ceci.stage import PipelineStage
from qp import Ensemble

from rail.core.data import DataHandle
from rail.core.stage import RailStage
from rail.estimation.estimator import CatEstimator


class PZFactory:
    """Factory class to provide a unified interface to
    rail p(z) estimation algorithms.
    """

    _stage_dict: dict[str, CatEstimator] = {}

    @classmethod
    def reset(cls) -> None:
        """Reset the dictionary of cached stage objects"""
        cls._stage_dict = {}

    @classmethod
    def build_stage_instance(
        cls,
        stage_name: str,
        stage_class: type[CatEstimator],
        model_path: str | dict,
        data_path: str = "none",
        **config_params: dict,
    ) -> CatEstimator:
        """Build and configure an estimator that can evalute
        p(z) given an input catalog

        Parameters
        ----------
        stage_name: str
            Name of the stage instance, used to construct output file name

        stage_class: type
            Python class for the stage

        model_path: str| dict
            Path to the model file used by this estimator, or the model itself

        data_path: str
            Path to the input data, defaults to 'none'

        config_params: dict
            Configuration parameters for the stage

        Returns
        -------
        stage_obj: CatEstimator
            Newly constructed and configured Estimator instance
        """
        stage_obj = stage_class.make_stage(
            name=stage_name, model=model_path, input=data_path, **config_params
        )
        cls._stage_dict[stage_name] = stage_obj
        return stage_obj

    @classmethod
    def build_cat_estimator_stage(
        cls,
        stage_name: str,
        class_name: str,
        module_name: str,
        model_path: str,
        data_path: str = "none",
        **config_params: dict,
    ) -> CatEstimator:
        """Build and configure an estimator that can evalute
        p(z) given an input catalog

        Parameters
        ----------
        stage_name: str
            Name of the stage instance, used to construct output file name

        class_name: str
            Name of the class, e.g., TrainZEstimator, used to find the class

        module_name: str
            Name of the python module that constains the class, used to import it

        model_path: str
            Path to the model file used by this estimator

        data_path: str
            Path to the input data, defaults to 'none'

        config_params: dict
            Configuration parameters for the stage

        Returns
        -------
        stage_obj: CatEstimator
            Newly constructed and configured Estimator instance
        """
        stage_class = PipelineStage.get_stage(class_name, module_name)
        return cls.build_stage_instance(
            stage_name, stage_class, model_path, data_path, **config_params
        )

    @classmethod
    def get_cat_estimator_stage(
        cls,
        stage_name: str,
    ) -> CatEstimator:
        """Return a cached p(z) estimator"""
        try:
            return cls._stage_dict[stage_name]
        except KeyError as msg:
            raise KeyError(
                f"Could not find stage named {stage_name}, did you build it?"
                f"Existing stages are: {list(cls._stage_dict.keys())}"
            ) from msg

    @staticmethod
    def run_cat_estimator_stage(
        stage_obj: CatEstimator,
        data_path: str,
    ) -> DataHandle:
        """Run a p(z) estimator on an input data file

        Parameters
        ----------
        stage_obj: CatEstimator
            Object that will do the estimation

        Returns
        -------
        data_handle: DataHandle
            Object that can give access to the data
        """
        RailStage.data_store.clear()
        handle = stage_obj.get_handle("input", path=data_path, allow_missing=True)
        return stage_obj.estimate(handle)

    @staticmethod
    def estimate_single_pz(
        stage_obj: CatEstimator,
        data_table: dict,
        input_size: int = 1,
    ) -> Ensemble:
        """Run a p(z) estimator on some objects

        Parameters
        ----------
        stage_obj: CatEstimator
            Object that will do the estimation

        data_table: dict
            Input data presented as dict of numpy arrays objects

        input_size: int
            Number of objects in the input table

        Returns
        -------
        pz : qp.Ensemble
            Output pz
        """
        RailStage.data_store.clear()
        if stage_obj.model is None:
            stage_obj.open_model(**stage_obj.config)
        stage_obj._input_length = input_size
        stage_obj._process_chunk(0, input_size, data_table, True)
        return stage_obj._output_handle.data



from __future__ import annotations

from typing import Any, TypeVar

import numpy as np

from rail.core import common_params

T = TypeVar("T", bound="CatalogConfigBase")


class CatalogConfigBase:
    """Class that wraps the settings of shared configuration
    parameters needed work with the particular column names
    in a given catalog type
    """

    _sub_classes_by_class: dict[str, type[CatalogConfigBase]] = {}
    _sub_classes: dict[str, type[CatalogConfigBase]] = {}

    tag: str | None = None
    bandlist: list[str] = []
    maglims: list[float] = []
    a_env: list[float] = []
    hdf5_groupname: str = ""
    band_template: str = ""
    band_err_template: str = ""
    filter_file_template: str = ""
    ref_band: str = ""
    redshift_col: str = ""
    object_id_col: str = ""
    lsst_err_band_replace: list[float] = [0.1, 0.1, 0.1, 0.1, 0.1, 0.1]
    replace_error_vals: list[float] = [0.1, 0.1, 0.1, 0.1, 0.1, 0.1]
    zp_errors: list[float] = [0.1, 0.1, 0.1, 0.1, 0.1, 0.1]
    _active_tag: str | None = None
    _active_class: type | None = None

    def __init_subclass__(cls, **kwargs: Any) -> None:
        assert cls.tag is not None
        cls._sub_classes_by_class[cls.__name__] = cls
        cls._sub_classes[cls.tag] = cls

    @classmethod
    def active_tag(cls) -> str | None:
        """Return the currently active tag"""
        return cls._active_tag

    @classmethod
    def active_class(cls: type[T]) -> type[T] | None:
        """Return the currently active class"""
        return cls._active_class

    @classmethod
    def subclasses_by_class(cls) -> dict[str, type[CatalogConfigBase]]:
        """Return the dict of all the sub-classes keyed by class name"""
        return cls._sub_classes_by_class

    @classmethod
    def subclasses(cls) -> dict[str, type[CatalogConfigBase]]:
        """Return the dict of all the sub-classes keyed by tag"""
        return cls._sub_classes

    @classmethod
    def get_class(cls, class_name: str, module_name: str) -> type[CatalogConfigBase]:
        """Return a class my name, loading it if needed"""
        if class_name not in cls._sub_classes_by_class:  # pragma: no cover
            __import__(module_name)
        return cls._sub_classes_by_class[class_name]

    @classmethod
    def apply_class(cls, class_name: str) -> None:
        """Activate a particular class"""
        the_class = cls._sub_classes_by_class[class_name]
        cls._active_tag = the_class.tag
        cls._active_class = the_class
        cls._active_class._apply()

    @classmethod
    def apply(cls, tag: str) -> None:
        """Activate a particular tag"""
        cls._active_tag = tag
        cls._active_class = cls._sub_classes[tag]
        cls._active_class._apply()

    @classmethod
    def _apply(cls) -> None:
        base_dict = cls.build_base_dict()
        common_params.set_param_defaults(**base_dict)
        cls._apply_hook()

    @classmethod
    def _apply_hook(cls) -> None:
        return

    @classmethod
    def band_name_dict(cls) -> dict[str, str]:
        """Retrun the mapping from band to column names"""
        return {band: cls.band_template.format(band=band) for band in cls.bandlist}

    @classmethod
    def _build_maglim_dict(cls) -> dict[str, float]:
        """Construct the mapping from band to limiting magnitude"""
        maglim_dict = {}
        for band, limx in zip(cls.bandlist, cls.maglims):
            maglim_dict[cls.band_template.format(band=band)] = limx
        return maglim_dict

    @classmethod
    def _build_a_env_dict(cls) -> dict[str, float]:
        """Construct the mapping from band to a_env"""
        a_env_dict = {}
        for band, limx in zip(cls.bandlist, cls.a_env):
            a_env_dict[cls.band_template.format(band=band)] = limx
        return a_env_dict

    @classmethod
    def _build_band_names(cls) -> list[str]:
        """Construct the list of column names for the filter magnitudes"""
        return [cls.band_template.format(band=band) for band in cls.bandlist]

    @classmethod
    def _build_band_err_names(cls) -> list[str]:
        """Construct the list of column names for uncertainties on the the filter magnitudes"""
        return [cls.band_err_template.format(band=band) for band in cls.bandlist]

    @classmethod
    def _build_err_dict(cls) -> dict[str, str| None]:
        """Construct the mapping from magnitude columns to associated uncertainties"""
        the_dict: dict[str, str|None] = {
            cls.band_template.format(band=band): cls.band_err_template.format(band=band)
            for band in cls.bandlist
        }
        the_dict[cls.redshift_col] = None
        return the_dict

    @classmethod
    def _build_ref_band(cls, ref_band: str = "i") -> str:
        """Contruct the name of the reference band"""
        return cls.band_template.format(band=ref_band)

    @classmethod
    def _build_filter_file_bandlist(cls) -> list[str]:
        """Contruct the list names used to find the filter files"""
        return [cls.filter_file_template.format(band=band) for band in cls.bandlist]

    @classmethod
    def build_base_dict(cls) -> dict:
        """Construct the dict of overrides for the shared paramters"""
        base_dict: dict = {}
        base_dict.update(
            bands=cls._build_band_names(),
            err_bands=cls._build_band_err_names(),
            err_dict=cls._build_err_dict(),
            mag_limits=cls._build_maglim_dict(),
            band_a_env=cls._build_a_env_dict(),
            ref_band=cls._build_ref_band(cls.ref_band),
            redshift_col=cls.redshift_col,
            nondetect_val=np.nan,
            nonobserved_val=np.inf,
            hdf5_groupname=cls.hdf5_groupname,
            replace_error_vals=cls.replace_error_vals,
            filter_list=cls._build_filter_file_bandlist(),
            zp_errors=cls.zp_errors,
        )
        return base_dict


class HscCatalogConfig(CatalogConfigBase):
    """Configuration for HSC data"""

    tag = "hsc"
    bandlist = ['g', 'r', 'i', 'z', 'y']
    maglims = [27.66, 27.25, 26.6, 26.24, 25.35]
    a_env = [3.64, 2.70, 2.06, 1.58, 1.31]
    band_template = "HSC{band}_cmodel_dered"
    band_err_template = "{band}_cmodel_magerr"
    filter_file_template = "DC2LSST_{band}"
    ref_band = "i"
    redshift_col = "specz_redshift"
    object_id_col = "object_id"
    hdf5_groupname = ""
    replace_error_vals = [0.1, 0.1, 0.1, 0.1, 0.1]
    zp_errors = [0.1, 0.1, 0.1, 0.1, 0.1]


class Dc2CatalogConfig(CatalogConfigBase):
    """Configuration for DC2 data"""

    tag = "dc2"
    bandlist = ['u', 'g', 'r', 'i', 'z', 'y']
    maglims = [24.0, 27.66, 27.25, 26.6, 26.24, 25.35]
    a_env = [4.81, 3.64, 2.70, 2.06, 1.58, 1.31]
    band_template = "mag_{band}_cModel_obj_dered"
    band_err_template = "magerr_{band}_cModel_obj"
    filter_file_template = "DC2LSST_{band}"
    ref_band = "i"
    redshift_col = "redshift_true"
    object_id_col = "objectId_obj"
    hdf5_groupname = ""
    replace_error_vals = [0.1, 0.1, 0.1, 0.1, 0.1, 0.1]


class RubinCatalogConfig(CatalogConfigBase):
    """Configuration for Rubin project data"""

    tag = "rubin"
    bandlist = ['u', 'g', 'r', 'i', 'z', 'y']
    maglims = [24.0, 27.66, 27.25, 26.6, 26.24, 25.35]
    a_env = [4.81, 3.64, 2.70, 2.06, 1.58, 1.31]
    band_template = "LSST_obs_{band}"
    band_err_template = "LSST_obs_{band}_err"
    filter_file_template = "DC2LSST_{band}"
    ref_band = "i"
    redshift_col = "true_redshift"
    object_id_col = "objectId"
    hdf5_groupname = ""
    replace_error_vals = [0.1, 0.1, 0.1, 0.1, 0.1, 0.1]


class Roman3BandCatalogConfig(CatalogConfigBase):
    """Configuration for Rubin data from Roman / Rubin simulations"""

    tag = "roman_3band"
    bandlist = ["Y106", "J129", "H158"]
    maglims = [26.4, 26.4, 26.4]
    a_env = [1.14025753, 0.83118224, 0.59966235]
    band_template = "ROMAN_obs_{band}"
    band_err_template = "ROMAN_obs_{band}_err"
    filter_file_template = "roman_{band}"
    ref_band = "Y106"
    redshift_col = "redshift"
    object_id_col = "objectId"
    hdf5_groupname = ""
    replace_error_vals = [0.1, 0.1, 0.1]
    zp_errors = [0.1, 0.1, 0.1]


class Roman7BandCatalogConfig(CatalogConfigBase):
    """Configuration for Rubin data from Roman / Rubin simulations"""

    tag = "roman_7band"
    bandlist = ["Z087", "Y106", "J129", "W146", "H158", "F184", "K213"]
    maglims = [27.4, 27.4, 27.4, 27.4, 27.4, 27.4, 27.4]
    a_env = [
        1.57491325,
        1.14025753,
        0.83118224,
        0.68098202,
        0.59966235,
        0.46923204,
        0.37072579,
    ]
    band_template = "ROMAN_obs_{band}"
    band_err_template = "ROMAN_obs_{band}_err"
    filter_file_template = "roman_{band}"
    ref_band = "Y106"
    redshift_col = "redshift"
    object_id_col = "objectId"
    hdf5_groupname = ""
    replace_error_vals = [0.1, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1]
    zp_errors = [0.1, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1]


class RomanRubinCatalogConfig(CatalogConfigBase):
    """Configuration for Rubin data from Roman / Rubin simulations"""

    tag = "roman_rubin"
    bandlist = ['u', 'g', 'r', 'i', 'z', 'y']
    maglims = [24.0, 27.66, 27.25, 26.6, 26.24, 25.35]
    a_env = [4.81, 3.64, 2.70, 2.06, 1.58, 1.31]
    band_template = "LSST_obs_{band}"
    band_err_template = "LSST_obs_{band}_err"
    filter_file_template = "DC2LSST_{band}"
    ref_band = "i"
    redshift_col = "redshift"
    object_id_col = "objectId"
    hdf5_groupname = ""
    replace_error_vals = [0.1, 0.1, 0.1, 0.1, 0.1, 0.1]


class ComCamCatalogConfig(CatalogConfigBase):
    """Configuration for ComCam data"""

    tag = "com_cam"
    bandlist = ['u', 'g', 'r', 'i', 'z', 'y']
    maglims = [26.4, 27.8, 27.1, 26.7, 25.8, 24.6]
    a_env = [4.81, 3.64, 2.70, 2.06, 1.58, 1.31]
    band_template = "{band}_cModelMag"
    band_err_template = "{band}_cModelMagErr"
    filter_file_template = "comcam_{band}"
    ref_band = "i"
    redshift_col = "redshift"


class RubinFLCatalogConfig(CatalogConfigBase):
    """Configuration for Rubin first look data """

    tag = "rubin_fl"
    bandlist = ['u', 'g', 'r', 'i']
    maglims = [26.4, 27.8, 27.1, 26.7]
    a_env = [4.81, 3.64, 2.70, 2.06]
    band_template = "{band}_cModelMag"
    band_err_template = "{band}_cModelMagErr"
    filter_file_template = "DC2LSST_{band}"
    ref_band = "i"
    redshift_col = "redshift"
    replace_error_vals = [0.1, 0.1, 0.1, 0.1]
    zp_errors = [0.1, 0.1, 0.1, 0.1]


class ComCamGaapCatalogConfig(CatalogConfigBase):
    """Configuration for ComCam data"""

    tag = "com_cam_gaap"
    bandlist = ['u', 'g', 'r', 'i', 'z', 'y']
    maglims = [26.4, 27.8, 27.1, 26.7, 25.8, 24.6]
    a_env = [4.81, 3.64, 2.70, 2.06, 1.58, 1.31]
    band_template = "{band}_gaap1p0Mag"
    band_err_template = "{band}_gaap1p0MagErr"
    filter_file_template = "comcam_{band}"
    ref_band = "i"
    redshift_col = "redshift"
    object_id_col = "objectId"
    hdf5_groupname = ""
    replace_error_vals = [0.1, 0.1, 0.1, 0.1, 0.1, 0.1]

        
class ComCamGaap4BandCatalogConfig(CatalogConfigBase):
    """Configuration for ComCam data"""

    tag = "com_cam_gaap_4band"
    bandlist = ['g', 'r', 'i', 'z']
    maglims = [27.8, 27.1, 26.7, 25.8]
    a_env = [3.64, 2.70, 2.06, 1.58]
    band_template = "{band}_gaap1p0Mag"
    band_err_template = "{band}_gaap1p0MagErr"
    filter_file_template = "comcam_{band}"
    ref_band = "i"
    redshift_col = "redshift"
    object_id_col = "objectId"
    hdf5_groupname = ""
    replace_error_vals = [ 0.1, 0.1, 0.1, 0.1]


class ComCamKronCatalogConfig(CatalogConfigBase):
    """Configuration for ComCam data"""

    tag = "com_cam_kron"
    bandlist = ['u', 'g', 'r', 'i', 'z', 'y']
    maglims = [26.4, 27.8, 27.1, 26.7, 25.8, 24.6]
    a_env = [4.81, 3.64, 2.70, 2.06, 1.58, 1.31]
    band_template = "{band}_kronMag"
    band_err_template = "{band}_kronMagErr"
    filter_file_template = "comcam_{band}"
    ref_band = "i"
    redshift_col = "redshift"
    object_id_col = "objectId"
    hdf5_groupname = ""
    replace_error_vals = [0.1, 0.1, 0.1, 0.1, 0.1, 0.1]


class ComCamEuclidCatalogConfig(CatalogConfigBase):
    """Configuration for ComCam data"""

    tag = "com_cam_euclid"
    bandlist = ['u', 'g', 'r', 'i', 'z', 'y']
    maglims = [26.4, 27.8, 27.1, 26.7, 25.8, 24.6]
    a_env = [4.81, 3.64, 2.70, 2.06, 1.58, 1.31]
    band_template = "{band}_gaap1p0Mag"
    band_err_template = "{band}_gaap1p0MagErr"
    filter_file_template = "comcam_{band}"
    ref_band = "i"
    redshift_col = "redshift"
    object_id_col = "objectId"
    hdf5_groupname = ""
    replace_error_vals = [0.1, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1]
    zp_errors = [0.1, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1]

    @classmethod
    def band_name_dict(cls) -> dict[str, str]:
        bands = super().band_name_dict()
        bands["vis"] = "euclid_vis_psfMag"
        bands["Y"] = "euclid_y_unifMag"
        bands["J"] = "euclid_j_unifMag"
        bands["H"] = "euclid_h_unifMag"
        return bands

    @classmethod
    def _build_maglim_dict(cls) -> dict[str, float]:
        maglim_dict = super()._build_maglim_dict()
        maglim_dict["euclid_vis_psfMag"] = 26.0
        maglim_dict["euclid_y_unifMag"] = 23.8
        maglim_dict["euclid_j_unifMag"] = 24.0
        maglim_dict["euclid_h_unifMag"] = 24.0
        return maglim_dict

    @classmethod
    def _build_a_env_dict(cls) -> dict[str, float]:
        a_env_dict = super()._build_a_env_dict()
        a_env_dict["euclid_vis_psfMag"] = 0.0
        a_env_dict["euclid_y_unifMag"] = 0.0
        a_env_dict["euclid_j_unifMag"] = 0.0
        a_env_dict["euclid_h_unifMag"] = 0.0
        return a_env_dict

    @classmethod
    def _build_band_names(cls) -> list[str]:
        bands = [cls.band_template.format(band=band) for band in cls.bandlist]
        bands += [
            "euclid_vis_psfMag",
            "euclid_y_unifMag",
            "euclid_j_unifMag",
            "euclid_h_unifMag",
        ]
        return bands

    @classmethod
    def _build_band_err_names(cls) -> list[str]:
        band_errs = [cls.band_err_template.format(band=band) for band in cls.bandlist]
        band_errs += [
            "euclid_vis_psfMagErr",
            "euclid_y_unifMagErr",
            "euclid_j_unifMagErr",
            "euclid_h_unifMagErr",
        ]
        return band_errs

    @classmethod
    def _build_err_dict(cls) -> dict[str, str | None]:
        the_dict = super()._build_err_dict()
        the_dict["euclid_vis_psfMag"] = "euclid_vis_psfMagErr"
        the_dict["euclid_y_unifMag"] = "euclid_y_unifMagErr"
        the_dict["euclid_j_unifMag"] = "euclid_j_unifMagErr"
        the_dict["euclid_h_unifMag"] = "euclid_h_unifMagErr"
        return the_dict

    @classmethod
    def _build_ref_band(cls, ref_band: str = "i") -> str:
        return cls.band_template.format(band=ref_band)

    @classmethod
    def _build_filter_file_bandlist(cls) -> list[str]:
        """Contruct the name of the reference band"""
        filter_list = [
            cls.filter_file_template.format(band=band) for band in cls.bandlist
        ]
        filter_list += [
            "euclid_vis",
            "euclid_y",
            "euclid_j",
            "euclid_h",
        ]
        return filter_list


class ComCamEuclidNIRCatalogConfig(CatalogConfigBase):
    """Configuration for ComCam data"""

    tag = "com_cam_euclid_nir"
    bandlist = ['u', 'g', 'r', 'i', 'z', 'y']
    maglims = [26.4, 27.8, 27.1, 26.7, 25.8, 24.6]
    a_env = [4.81, 3.64, 2.70, 2.06, 1.58, 1.31]
    band_template = "{band}_gaap1p0Mag"
    band_err_template = "{band}_gaap1p0MagErr"
    filter_file_template = "comcam_{band}"
    ref_band = "i"
    redshift_col = "redshift"
    object_id_col = "objectId"
    hdf5_groupname = ""
    replace_error_vals = [0.1, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1]
    zp_errors = [0.1, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1]

    @classmethod
    def band_name_dict(cls) -> dict[str, str]:
        bands = super().band_name_dict()
        bands["Y"] = "euclid_y_unifMag"
        bands["J"] = "euclid_j_unifMag"
        bands["H"] = "euclid_h_unifMag"
        return bands

    @classmethod
    def _build_maglim_dict(cls) -> dict[str, float]:
        maglim_dict = super()._build_maglim_dict()
        maglim_dict["euclid_y_unifMag"] = 23.8
        maglim_dict["euclid_j_unifMag"] = 24.0
        maglim_dict["euclid_h_unifMag"] = 24.0
        return maglim_dict

    @classmethod
    def _build_a_env_dict(cls) -> dict[str, float]:
        a_env_dict = super()._build_a_env_dict()
        a_env_dict["euclid_y_unifMag"] = 0.0
        a_env_dict["euclid_j_unifMag"] = 0.0
        a_env_dict["euclid_h_unifMag"] = 0.0
        return a_env_dict

    @classmethod
    def _build_band_names(cls) -> list[str]:
        bands = [cls.band_template.format(band=band) for band in cls.bandlist]
        bands += [
            "euclid_y_unifMag",
            "euclid_j_unifMag",
            "euclid_h_unifMag",
        ]
        return bands

    @classmethod
    def _build_band_err_names(cls) -> list[str]:
        band_errs = [cls.band_err_template.format(band=band) for band in cls.bandlist]
        band_errs += [
            "euclid_y_unifMagErr",
            "euclid_j_unifMagErr",
            "euclid_h_unifMagErr",
        ]
        return band_errs

    @classmethod
    def _build_err_dict(cls) -> dict[str, str | None]:
        the_dict = super()._build_err_dict()
        the_dict["euclid_y_unifMag"] = "euclid_y_unifMagErr"
        the_dict["euclid_j_unifMag"] = "euclid_j_unifMagErr"
        the_dict["euclid_h_unifMag"] = "euclid_h_unifMagErr"
        return the_dict

    @classmethod
    def _build_ref_band(cls, ref_band: str = "i") -> str:
        return cls.band_template.format(band=ref_band)

    @classmethod
    def _build_filter_file_bandlist(cls) -> list[str]:
        """Contruct the name of the reference band"""
        filter_list = [
            cls.filter_file_template.format(band=band) for band in cls.bandlist
        ]
        filter_list += [
            "euclid_y",
            "euclid_j",
            "euclid_h",
        ]
        return filter_list

class RomanPlusRubinCatalogConfig(CatalogConfigBase):
    """Configuration for Roman + Rubin bands in Roman / Rubin simulations"""

    tag = "roman_plus_rubin"
    bandlist = ['u', 'g', 'r', 'i', 'z', 'y']
    maglims = [24.0, 27.66, 27.25, 26.6, 26.24, 25.35]
    a_env = [4.81, 3.64, 2.70, 2.06, 1.58, 1.31]
    band_template = "LSST_obs_{band}"
    band_err_template = "LSST_obs_{band}_err"
    filter_file_template = "DC2LSST_{band}"
    ref_band = "i"
    redshift_col = "redshift"
    object_id_col = "objectId"
    hdf5_groupname = ""
    replace_error_vals = [0.1, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1]

    @classmethod
    def band_name_dict(cls) -> dict[str, str]:
        bands = super().band_name_dict()
        bands["Y"] = "ROMAN_obs_Y106"
        bands["J"] = "ROMAN_obs_J129"
        bands["H"] = "ROMAN_obs_H158"
        bands["F"] = "ROMAN_obs_F184"
        return bands

    @classmethod
    def _build_maglim_dict(cls) -> dict[str, float]:
        maglim_dict = super()._build_maglim_dict()
        maglim_dict["ROMAN_obs_Y106"] = 27.6
        maglim_dict["ROMAN_obs_J129"] = 27.8
        maglim_dict["ROMAN_obs_H158"] = 28.1
        maglim_dict["ROMAN_obs_F184"] = 27.5
        return maglim_dict

    @classmethod
    def _build_a_env_dict(cls) -> dict[str, float]:
        a_env_dict = super()._build_a_env_dict()
        a_env_dict["ROMAN_obs_Y106"] = 0.83118224
        a_env_dict["ROMAN_obs_J129"] = 0.68098202
        a_env_dict["ROMAN_obs_H158"] = 0.59966235
        a_env_dict["ROMAN_obs_F184"] = 0.46923204
        return a_env_dict

    @classmethod
    def _build_band_names(cls) -> list[str]:
        bands = [cls.band_template.format(band=band) for band in cls.bandlist]
        bands += [
            "ROMAN_obs_Y106",
            "ROMAN_obs_J129",
            "ROMAN_obs_H158",
            "ROMAN_obs_F184",
        ]
        return bands

    @classmethod
    def _build_band_err_names(cls) -> list[str]:
        band_errs = [cls.band_err_template.format(band=band) for band in cls.bandlist]
        band_errs += [
            "ROMAN_obs_Y106_err",
            "ROMAN_obs_J129_err",
            "ROMAN_obs_H158_err",
            "ROMAN_obs_F184_err",
        ]
        return band_errs

    @classmethod
    def _build_err_dict(cls) -> dict[str, str | None]:
        the_dict = super()._build_err_dict()
        the_dict["ROMAN_obs_Y106"] = "ROMAN_obs_Y106_err"
        the_dict["ROMAN_obs_J129"] = "ROMAN_obs_J129_err"
        the_dict["ROMAN_obs_H158"] = "ROMAN_obs_H158_err"
        the_dict["ROMAN_obs_F184"] = "ROMAN_obs_F184_err"
        return the_dict

    @classmethod
    def _build_ref_band(cls, ref_band: str = "i") -> str:
        return cls.band_template.format(band=ref_band)

    @classmethod
    def _build_filter_file_bandlist(cls) -> list[str]:
        """Contruct the name of the reference band"""
        filter_list = [
            cls.filter_file_template.format(band=band) for band in cls.bandlist
        ]
        filter_list += [
            "ROMAN_obs_Y106",
            "ROMAN_obs_J129",
            "ROMAN_obs_H158",
            "ROMAN_obs_F184",
        ]
        return filter_list


class Roman3BandPlusRubinCatalogConfig(CatalogConfigBase):
    """Configuration for Roman3Band + Rubin bands in Roman / Rubin simulations"""

    tag = "roman_3band_rubin"
    bandlist = ['u', 'g', 'r', 'i', 'z', 'y']
    maglims = [24.0, 27.66, 27.25, 26.6, 26.24, 25.35]
    a_env = [4.81, 3.64, 2.70, 2.06, 1.58, 1.31]
    band_template = "LSST_obs_{band}"
    band_err_template = "LSST_obs_{band}_err"
    filter_file_template = "DC2LSST_{band}"
    ref_band = "i"
    redshift_col = "redshift"
    object_id_col = "objectId"
    hdf5_groupname = ""
    replace_error_vals = [0.1, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1]
    zp_errors = [0.1, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1]

    @classmethod
    def band_name_dict(cls) -> dict[str, str]:
        bands = super().band_name_dict()
        bands["Y"] = "ROMAN_obs_Y106"
        bands["J"] = "ROMAN_obs_J129"
        bands["H"] = "ROMAN_obs_H158"
        return bands

    @classmethod
    def _build_maglim_dict(cls) -> dict[str, float]:
        maglim_dict = super()._build_maglim_dict()
        maglim_dict["ROMAN_obs_Y106"] = 26.5
        maglim_dict["ROMAN_obs_J129"] = 26.4
        maglim_dict["ROMAN_obs_H158"] = 26.4
        return maglim_dict

    @classmethod
    def _build_a_env_dict(cls) -> dict[str, float]:
        a_env_dict = super()._build_a_env_dict()
        a_env_dict["ROMAN_obs_Y106"] = 0.83118224
        a_env_dict["ROMAN_obs_J129"] = 0.68098202
        a_env_dict["ROMAN_obs_H158"] = 0.59966235
        return a_env_dict

    @classmethod
    def _build_band_names(cls) -> list[str]:
        bands = [cls.band_template.format(band=band) for band in cls.bandlist]
        bands += [
            "ROMAN_obs_Y106",
            "ROMAN_obs_J129",
            "ROMAN_obs_H158",
        ]
        return bands

    @classmethod
    def _build_band_err_names(cls) -> list[str]:
        band_errs = [cls.band_err_template.format(band=band) for band in cls.bandlist]
        band_errs += [
            "ROMAN_obs_Y106_err",
            "ROMAN_obs_J129_err",
            "ROMAN_obs_H158_err",
        ]
        return band_errs

    @classmethod
    def _build_err_dict(cls) -> dict[str, str | None]:
        the_dict = super()._build_err_dict()
        the_dict["ROMAN_obs_Y106"] = "ROMAN_obs_Y106_err"
        the_dict["ROMAN_obs_J129"] = "ROMAN_obs_J129_err"
        the_dict["ROMAN_obs_H158"] = "ROMAN_obs_H158_err"
        return the_dict

    @classmethod
    def _build_ref_band(cls, ref_band: str = "i") -> str:
        return cls.band_template.format(band=ref_band)

    @classmethod
    def _build_filter_file_bandlist(cls) -> list[str]:
        """Contruct the name of the reference band"""
        filter_list = [
            cls.filter_file_template.format(band=band) for band in cls.bandlist
        ]
        filter_list += [
            "roman_Y106",
            "roman_J129",
            "roman_H158",
        ]
        return filter_list


class Roman7BandPlusRubinCatalogConfig(CatalogConfigBase):
    """Configuration for Roman3Band + Rubin bands in Roman / Rubin simulations"""

    tag = "roman_7band_rubin"
    bandlist = ['u', 'g', 'r', 'i', 'z', 'y']
    maglims = [24.0, 27.66, 27.25, 26.6, 26.24, 25.35]
    a_env = [4.81, 3.64, 2.70, 2.06, 1.58, 1.31]
    band_template = "LSST_obs_{band}"
    band_err_template = "LSST_obs_{band}_err"
    filter_file_template = "DC2LSST_{band}"
    ref_band = "i"
    redshift_col = "redshift"
    object_id_col = "objectId"
    hdf5_groupname = ""
    replace_error_vals = [
        0.1,
        0.1,
        0.1,
        0.1,
        0.1,
        0.1,
        0.1,
        0.1,
        0.1,
        0.1,
        0.1,
        0.1,
        0.1,
    ]
    zp_errors = [0.1, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1]

    @classmethod
    def band_name_dict(cls) -> dict[str, str]:
        bands = super().band_name_dict()

        bands["Z"] = "ROMAN_obs_Z087"
        bands["Y"] = "ROMAN_obs_Y106"
        bands["J"] = "ROMAN_obs_J129"
        bands["W"] = "ROMAN_obs_W146"
        bands["H"] = "ROMAN_obs_H158"
        bands["F"] = "ROMAN_obs_F184"
        bands["K"] = "ROMAN_obs_K213"
        return bands

    @classmethod
    def _build_maglim_dict(cls) -> dict[str, float]:
        maglim_dict = super()._build_maglim_dict()
        maglim_dict["ROMAN_obs_Z087"] = 27.7
        maglim_dict["ROMAN_obs_Y106"] = 27.7
        maglim_dict["ROMAN_obs_J129"] = 27.6
        maglim_dict["ROMAN_obs_W146"] = 28.3
        maglim_dict["ROMAN_obs_H158"] = 27.5
        maglim_dict["ROMAN_obs_F184"] = 27.0
        maglim_dict["ROMAN_obs_K213"] = 25.9
        return maglim_dict

    @classmethod
    def _build_a_env_dict(cls) -> dict[str, float]:
        a_env_dict = super()._build_a_env_dict()
        a_env_dict["ROMAN_obs_Z087"] = 1.57491325
        a_env_dict["ROMAN_obs_Y106"] = 1.14025753
        a_env_dict["ROMAN_obs_J129"] = 0.83118224
        a_env_dict["ROMAN_obs_W146"] = 0.68098202
        a_env_dict["ROMAN_obs_H158"] = 0.59966235
        a_env_dict["ROMAN_obs_F184"] = 0.46923204
        a_env_dict["ROMAN_obs_K213"] = 0.37072579
        return a_env_dict

    @classmethod
    def _build_band_names(cls) -> list[str]:
        bands = [cls.band_template.format(band=band) for band in cls.bandlist]
        bands += [
            "ROMAN_obs_Z087",
            "ROMAN_obs_Y106",
            "ROMAN_obs_J129",
            "ROMAN_obs_W146",
            "ROMAN_obs_H158",
            "ROMAN_obs_F184",
            "ROMAN_obs_K213",
        ]
        return bands

    @classmethod
    def _build_band_err_names(cls) -> list[str]:
        band_errs = [cls.band_err_template.format(band=band) for band in cls.bandlist]
        band_errs += [
            "ROMAN_obs_Z087_err",
            "ROMAN_obs_Y106_err",
            "ROMAN_obs_J129_err",
            "ROMAN_obs_W146_err",
            "ROMAN_obs_H158_err",
            "ROMAN_obs_F184_err",
            "ROMAN_obs_K213_err",
        ]
        return band_errs

    @classmethod
    def _build_err_dict(cls) -> dict[str, str | None]:
        the_dict = super()._build_err_dict()
        the_dict["ROMAN_obs_Z087"] = "ROMAN_obs_Z087_err"
        the_dict["ROMAN_obs_Y106"] = "ROMAN_obs_Y106_err"
        the_dict["ROMAN_obs_J129"] = "ROMAN_obs_J129_err"
        the_dict["ROMAN_obs_W146"] = "ROMAN_obs_W146_err"
        the_dict["ROMAN_obs_H158"] = "ROMAN_obs_H158_err"
        the_dict["ROMAN_obs_F184"] = "ROMAN_obs_F184_err"
        the_dict["ROMAN_obs_K213"] = "ROMAN_obs_K213_err"
        return the_dict

    @classmethod
    def _build_ref_band(cls, ref_band: str = "i") -> str:
        return cls.band_template.format(band=ref_band)

    @classmethod
    def _build_filter_file_bandlist(cls) -> list[str]:
        """Contruct the name of the reference band"""
        filter_list = [
            cls.filter_file_template.format(band=band) for band in cls.bandlist
        ]
        filter_list += [
            "roman_Z087",
            "roman_Y106",
            "roman_J129",
            "roman_W146",
            "roman_H158",
            "roman_F184",
            "roman_K213",
        ]
        return filter_list



DP1_FILTERS = 'ugrizy'
DP1_FLUX_TYPES = ['cModel', 'kron', 'sersic', 'gaap1p0', 'psf', 'gaap3p0']
DP1_A_ENV_IN = Dc2CatalogConfig.build_base_dict()['band_a_env']
DP1_MAGLIMS_IN = Dc2CatalogConfig.build_base_dict()['mag_limits']
DP1_A_ENV_LIST = []
DP1_REPLACE_ERR_VALS = []
DP1_MAGLIMS = []
DP1_ZP_ERRORS = []
DP1_BANDS = []
for filter_ in DP1_FILTERS:
    for flux_type_ in DP1_FLUX_TYPES:
        band = f"{filter_}_{flux_type_}"
        DP1_BANDS.append(band)
        DP1_A_ENV_LIST.append(DP1_A_ENV_IN[f"mag_{filter_}_cModel_obj_dered"])
        DP1_MAGLIMS.append(DP1_MAGLIMS_IN[f"mag_{filter_}_cModel_obj_dered"])
        DP1_REPLACE_ERR_VALS.append(0.1)
        DP1_ZP_ERRORS.append(0.1)


class DP1AllFluxesCatalogConfig(CatalogConfigBase):
    """Configuration for ComCam data"""

    tag = "dp1_all"
    bandlist = DP1_BANDS
    maglims = DP1_MAGLIMS
    a_env = DP1_A_ENV_LIST
    band_template = "{band}Mag"
    band_err_template = "{band}MagErr"
    filter_file_template = "DC2LSST_{band}"
    ref_band = "i_psf"
    redshift_col = "redshift"
    object_id_col = "objectId"
    hdf5_groupname = ""
    replace_error_vals = DP1_REPLACE_ERR_VALS
    zp_errors = DP1_ZP_ERRORS




apply_defaults = CatalogConfigBase.apply



import numpy as np

from .base import MetricEvaluator


class PointStatsEz(MetricEvaluator):
    """Copied from PZDC1paper repo. Adapted to remove the cut based on
    magnitude."""

    def __init__(self, pzvec: np.ndarray, szvec: np.ndarray) -> None:
        """An object that takes in the vectors of the point photo-z
        the spec-z, and the i-band magnitudes for calculating the
        point statistics

        Calculates:
        ez = (pz - sz) / (1 + sz), the quantity will be useful for calculating
        statistics


        Parameters
        ----------
        pzvec : ndarray
            Array of the point photo-z values
        szvec : ndarray
            array of the spec-z values
        """
        super().__init__(None)
        self.pzs = pzvec
        self.szs = szvec
        ez = (pzvec - szvec) / (1.0 + szvec)
        self.ez = ez

    def evaluate(self) -> float | np.ndarray:
        """Return the ez values"""
        return self.ez


class PointSigmaIQR(PointStatsEz):
    """Calculate sigmaIQR"""

    def evaluate(self) -> float:
        """Calculate the width of the e_z distribution
        using the Interquartile range

        Returns
        -------
        ``sigma_IQR`` float. Width of ez distribution for full sample
        """
        x75, x25 = np.percentile(self.ez, [75.0, 25.0])
        iqr = x75 - x25
        sigma_iqr = iqr / 1.349
        return sigma_iqr


class PointBias(PointStatsEz):
    """calculates the bias of the ez and ez_magcut samples.

    In keeping with the Science Book, this is just the median of the ez values
    """

    def evaluate(self) -> float:
        """
        Returns
        -------
        ``bias`` ndarray. Median of the full ez sample
        """
        return np.median(self.ez)


class PointOutlierRate(PointStatsEz):
    """Calculates the catastrophic outlier rate, defined in the
    Science Book as the number of galaxies with ez larger than
    max(0.06,3sigma).  This keeps the fraction reasonable when
    sigma is very small.
    """

    def evaluate(self) -> float:
        """
        Returns
        -------
        ``frac`` float. Fraction of catastrophic outliers for full sample
        """
        num = len(self.ez)
        sig_iqr = PointSigmaIQR(self.pzs, self.szs).evaluate()
        threesig = 3.0 * sig_iqr
        cutcriterion = np.maximum(0.06, threesig)
        mask = np.fabs(self.ez) > cutcriterion
        outlier = np.sum(mask)
        frac = float(outlier) / float(num)
        return frac


class PointSigmaMAD(PointStatsEz):
    """Function to calculate median absolute deviation and sigma
    based on MAD (just scaled up by 1.4826) for the full and
    magnitude trimmed samples of ez values
    """

    def evaluate(self) -> float:
        """
        Returns
        -------
        ``sigma_mad`` float. Sigma median absolute deviation for full sample.
        """
        tmpmed = np.median(self.ez)
        tmpx = np.fabs(self.ez - tmpmed)
        mad = np.median(tmpx)
        sigma_mad = mad * 1.4826
        return sigma_mad



from ceci.stage import PipelineStage

from rail.core.data import DataHandle
from rail.core.stage import RailStage


class ToolFactory:
    """Factory class to provide a unified interface to
    rail tool stages
    """

    _stage_dict: dict[str, RailStage] = {}

    @classmethod
    def reset(cls) -> None:
        """Reset the dictionary of cached stage objects"""
        cls._stage_dict = {}

    @classmethod
    def build_stage_instance(
        cls,
        stage_name: str,
        stage_class: type[RailStage],
        data_path: str = "none",
        **config_params: dict,
    ) -> RailStage:
        """Build and configure a rail tool stage

        Parameters
        ----------
        stage_name: str
            Name of the stage instance, used to construct output file name

        stage_class: type
            Python class for the stage

        data_path: str
            Path to the input data, defaults to 'none'

        config_params: dict
            Configuration parameters for the stage

        Returns
        -------
        stage_obj: RailStage
            Newly constructed and configured RailStage instance
        """
        stage_obj = stage_class.make_stage(
            name=stage_name, input=data_path, **config_params
        )
        cls._stage_dict[stage_name] = stage_obj
        return stage_obj

    @classmethod
    def build_tool_stage(
        cls,
        stage_name: str,
        class_name: str,
        module_name: str,
        data_path: str = "none",
        **config_params: dict,
    ) -> RailStage:
        """Build and configure a rail tool stage

        Parameters
        ----------
        stage_name: str
            Name of the stage instance, used to construct output file name

        class_name: str
            Name of the class, used to find the class

        module_name: str
            Name of the python module that constains the class, used to import it

        data_path: str
            Path to the input data, defaults to 'none'

        config_params: dict
            Configuration parameters for the stage

        Returns
        -------
        stage_obj: CatEstimator
            Newly constructed and configured Estimator instance
        """
        stage_class = PipelineStage.get_stage(class_name, module_name)
        return cls.build_stage_instance(
            stage_name, stage_class, data_path, **config_params
        )

    @classmethod
    def get_tool_stage(
        cls,
        stage_name: str,
    ) -> RailStage:
        """Return a tool stage"""
        try:
            return cls._stage_dict[stage_name]
        except KeyError as msg:
            raise KeyError(
                f"Could not find stage named {stage_name}, did you build it?"
                f"Existing stages are: {list(cls._stage_dict.keys())}"
            ) from msg

    @staticmethod
    def run_tool_stage(
        stage_obj: RailStage,
        data_path: str,
    ) -> DataHandle:
        """Run a p(z) estimator on an input data file

        Parameters
        ----------
        stage_obj: CatEstimator
            Object that will do the estimation

        Returns
        -------
        data_handle: DataHandle
            Object that can give access to the data
        """
        RailStage.data_store.clear()
        handle = stage_obj.get_handle("input", path=data_path, allow_missing=True)
        return stage_obj(handle)



import os

import qp

from rail.core.data import TableHandle
from rail.core.stage import RailStage
from rail.estimation.algos.true_nz import TrueNZHistogrammer

DS = RailStage.data_store
DS.__class__.allow_overwrite = True

true_nz_file = "src/rail/examples_data/testdata/validation_10gal.hdf5"
tomo_file = "src/rail/examples_data/testdata/output_tomo.hdf5"


def test_true_nz() -> None:
    DS.clear()
    true_nz = DS.read_file("true_nz", path=true_nz_file, handle_class=TableHandle)
    tomo_bins = DS.read_file("tomo_bins", path=tomo_file, handle_class=TableHandle)

    nz_hist = TrueNZHistogrammer.make_stage(
        name="true_nz",
        hdf5_groupname="photometry",
        redshift_col="redshift",
        zmin=0.0,
        zmax=3.0,
        nzbins=301,
    )
    out_hist = nz_hist.histogram(true_nz, tomo_bins)

    check_ens = qp.read(out_hist.path)
    assert check_ens.ancil["n_total"][0] == 10

    check_vals = [0, 5, 0, 0, 0]

    for i in range(5):
        nz_hist = TrueNZHistogrammer.make_stage(
            name="true_nz",
            hdf5_groupname="photometry",
            redshift_col="redshift",
            zmin=0.0,
            zmax=3.0,
            nzbins=301,
            selected_bin=i,
        )
        out_hist = nz_hist.histogram(true_nz, tomo_bins)
        check_ens = qp.read(out_hist.path)
        assert check_ens.ancil["n_total"][0] == check_vals[i]

        os.remove(
            nz_hist.get_output(nz_hist.get_aliased_tag("true_NZ"), final_name=True)
        )



import os
import pickle
from types import GeneratorType

import numpy as np
import pytest

from rail.core.common_params import copy_param, set_param_default
from rail.core.data import (
    DataHandle,
    DataStore,
    FitsHandle,
    Hdf5Handle,
    ModelHandle,
    PqHandle,
    QPHandle,
    QPOrTableHandle,
)
from rail.core.model import Model
from rail.core.stage import RailStage
from rail.utils.catalog_utils import CatalogConfigBase, RomanPlusRubinCatalogConfig
from rail.utils.path_utils import RAILDIR

# def test_data_file():
#    with pytest.raises(ValueError) as errinfo:
#        df = DataFile('dummy', 'x')


def do_data_handle(datapath: str, handle_class: type[DataHandle]) -> DataHandle:
    _DS = RailStage.data_store

    th = handle_class("data", path=datapath)

    with pytest.raises(ValueError) as _errinfo:
        th.write()

    assert handle_class.__name__ in DataHandle.get_sub_classes()

    assert not th.has_data
    try:
        _check_size = th.size()
    except NotImplementedError as msg:
        if not isinstance(th, FitsHandle):
            raise NotImplementedError(msg) from msg

    with pytest.raises(ValueError) as _errinfo:
        th.write_chunk(0, 1)
    assert th.has_path
    assert th.is_written
    data = th.read()
    data2 = th.read()

    assert data is data2
    assert th.has_data
    assert th.make_name("data") == f"data.{handle_class.suffix}"

    th2 = handle_class("data2", data=data)
    assert th2.has_data
    assert th2.size() > 0

    assert not th2.has_path
    assert not th2.is_written
    with pytest.raises(ValueError) as _errinfo:
        th2.open()
    with pytest.raises(ValueError) as _errinfo:
        th2.write()
    with pytest.raises(ValueError) as _errinfo:
        th2.write_chunk(0, 1)

    assert th2.make_name("data2") == f"data2.{handle_class.suffix}"
    assert str(th)
    assert str(th2)
    return th


def test_pq_handle() -> None:
    datapath = os.path.join(
        RAILDIR, "rail", "examples_data", "testdata", "test_dc2_training_9816.pq"
    )
    handle = do_data_handle(datapath, PqHandle)
    pqfile = handle.open()
    assert pqfile
    assert handle.fileObj is not None
    handle.close()
    assert handle.fileObj is None


def test_qp_handle() -> None:
    datapath = os.path.join(
        RAILDIR, "rail", "examples_data", "testdata", "output_BPZ_lite.hdf5"
    )
    handle = do_data_handle(datapath, QPHandle)
    qpfile = handle.open()
    assert qpfile
    assert handle.fileObj is not None
    handle.close()
    assert handle.fileObj is None

    with pytest.raises(TypeError) as _errInfo:
        _bad_dh = QPHandle(tag="bad_tag", data="this is not an Ensemble")


def test_qp_or_table_handle_qp() -> None:
    datapath = os.path.join(
        RAILDIR, "rail", "examples_data", "testdata", "output_BPZ_lite.hdf5"
    )
    handle = do_data_handle(datapath, QPOrTableHandle)
    qpfile = handle.open()
    assert qpfile
    assert handle.fileObj is not None
    handle.close()
    assert handle.fileObj is None

    x = handle.iterator(chunk_size=100)

    assert isinstance(x, GeneratorType)
    for i, xx in enumerate(x):
        assert xx[0] == i * 100
        assert xx[1] - xx[0] <= 100

    handle2 = QPOrTableHandle(tag="qp_or_table_qp_2", path=datapath)

    x2 = handle2.iterator(chunk_size=100)

    assert isinstance(x2, GeneratorType)
    for i2, xx2 in enumerate(x2):
        assert xx2[0] == i2 * 100
        assert xx2[1] - xx2[0] <= 100


def test_qp_or_table_handle_table() -> None:
    datapath = os.path.join(
        RAILDIR, "rail", "examples_data", "testdata", "test_dc2_training_9816.hdf5"
    )
    handle = do_data_handle(datapath, QPOrTableHandle)
    tablefile = handle.open()
    assert tablefile
    assert handle.fileObj is not None
    handle.close()
    assert handle.fileObj is None

    x = handle.iterator(chunk_size=100)

    assert isinstance(x, GeneratorType)
    for i, xx in enumerate(x):
        assert xx[0] == i * 100
        assert xx[1] - xx[0] <= 100

    handle2 = QPOrTableHandle(tag="qp_or_table_table_2", path=datapath)

    x2 = handle2.iterator(groupname="photometry", chunk_size=100)

    assert isinstance(x2, GeneratorType)
    for i2, xx2 in enumerate(x2):
        assert xx2[0] == i2 * 100
        assert xx2[1] - xx2[0] <= 100


def test_hdf5_handle() -> None:
    datapath = os.path.join(
        RAILDIR, "rail", "examples_data", "testdata", "test_dc2_training_9816.hdf5"
    )
    handle = do_data_handle(datapath, Hdf5Handle)
    with handle.open(mode="r") as f:
        assert f
        assert handle.fileObj is not None
    datapath_chunked = os.path.join(
        RAILDIR,
        "rail",
        "examples_data",
        "testdata",
        "test_dc2_training_9816_chunked.hdf5",
    )
    handle_chunked = Hdf5Handle("chunked", handle.data, path=datapath_chunked)
    from tables_io.utils.array_utils import (  # pylint: disable=import-outside-toplevel
        get_initialization_for_ODict,
        slice_dict,
    )

    num_rows = len(handle.data["photometry"]["id"])
    check_num_rows = len(handle()["photometry"]["id"])
    assert num_rows == check_num_rows
    chunk_size = 1000
    data = handle.data["photometry"]
    init_dict = get_initialization_for_ODict(data)
    with handle_chunked.open(mode="w") as fout:
        for k, v in init_dict.items():
            fout.create_dataset(k, v[0], v[1])
        for i in range(0, num_rows, chunk_size):
            start = i
            end = i + chunk_size
            end = min(end, num_rows)
            handle_chunked.set_data(
                slice_dict(handle.data["photometry"], slice(start, end)), partial=True
            )
            handle_chunked.write_chunk(start, end)
    write_size = handle_chunked.size()
    assert len(handle_chunked.data) <= 1000
    data_called = handle_chunked()
    assert len(data_called["id"]) == write_size
    read_chunked = Hdf5Handle("read_chunked", None, path=datapath_chunked)
    data_check = read_chunked.read()
    assert np.allclose(data["id"], data_check["id"])
    assert np.allclose(data_called["id"], data_check["id"])
    os.remove(datapath_chunked)


def test_fits_handle() -> None:
    datapath = os.path.join(
        RAILDIR, "rail", "examples_data", "testdata", "output_BPZ_lite.fits"
    )
    handle = do_data_handle(datapath, FitsHandle)
    fitsfile = handle.open()
    assert fitsfile
    assert handle.fileObj is not None
    handle.close()
    assert handle.fileObj is None


def test_model_handle() -> None:
    DS = RailStage.data_store
    DS.clear()
    model_path = os.path.join(
        RAILDIR,
        "rail",
        "examples_data",
        "estimation_data",
        "data",
        "CWW_HDFN_prior.pkl",
    )
    model_path_copy = os.path.join(
        RAILDIR,
        "rail",
        "examples_data",
        "estimation_data",
        "data",
        "CWW_HDFN_prior_copy.pkl",
    )
    model_path_wrap = os.path.join(
        RAILDIR,
        "rail",
        "examples_data",
        "estimation_data",
        "data",
        "CWW_HDFN_prior_wrap.pkl",
    )
    mh = ModelHandle("model", path=model_path)
    mh2 = ModelHandle("model2", path=model_path)
    Model.wrap(model_path, model_path_wrap)
    mh4 = ModelHandle("modelWrap", path=model_path_wrap)

    model1 = mh.read()
    model2 = mh2.read()

    model3 = mh.open()

    _model4 = mh4.read()

    assert model1 is model2
    assert model2 is model3

    mh3 = ModelHandle("model3", path=model_path_copy, data=model1)
    with mh3.open(mode="w") as fout:
        pickle.dump(obj=mh3.data, file=fout, protocol=pickle.HIGHEST_PROTOCOL)
    os.remove(model_path_copy)
    os.remove(model_path_wrap)


def test_data_store() -> None:
    DS = RailStage.data_store
    DS.clear()
    DS.__class__.allow_overwrite = False

    datapath_hdf5 = os.path.join(
        RAILDIR, "rail", "examples_data", "testdata", "test_dc2_training_9816.hdf5"
    )
    datapath_pq = os.path.join(
        RAILDIR, "rail", "examples_data", "testdata", "test_dc2_training_9816.pq"
    )
    datapath_hdf5_copy = os.path.join(
        RAILDIR, "rail", "examples_data", "testdata", "test_dc2_training_9816_copy.hdf5"
    )
    datapath_pq_copy = os.path.join(
        RAILDIR, "rail", "examples_data", "testdata", "test_dc2_training_9816_copy.pq"
    )

    DS.add_data("hdf5", None, Hdf5Handle, path=datapath_hdf5)
    DS.add_data("pq", None, PqHandle, path=datapath_pq)

    with DS.open("hdf5") as f:
        assert f

    data_pq = DS.read("pq")
    data_hdf5 = DS.read("hdf5")

    DS.add_data("pq_copy", data_pq, PqHandle, path=datapath_pq_copy)
    DS.add_data("hdf5_copy", data_hdf5, Hdf5Handle, path=datapath_hdf5_copy)
    DS.write("pq_copy")
    DS.write("hdf5_copy")

    with pytest.raises(KeyError) as _errinfo:
        DS.read("nope")
    with pytest.raises(KeyError) as _errinfo:
        DS.open("nope")
    with pytest.raises(KeyError) as _errinfo:
        DS.write("nope")

    with pytest.raises(TypeError) as _errinfo:
        DS["nope"] = None
    with pytest.raises(ValueError) as _errinfo:
        DS["pq"] = DS["pq"]
    with pytest.raises(ValueError) as _errinfo:
        DS.pq = DS["pq"]

    a_handle = DS.add_handle("pq_copy_2", PqHandle, path=datapath_pq_copy)
    assert a_handle

    assert repr(DS)

    DS2 = DataStore(pq=DS.pq)
    assert isinstance(DS2.pq, DataHandle)

    # pop the 'pq' data item to avoid overwriting file under git control
    DS.pop("pq")
    # pop the 'pq_copy_2' because it is empty
    DS.pop("pq_copy_2")

    DS.write_all()
    DS.write_all(force=True)

    os.remove(datapath_hdf5_copy)
    os.remove(datapath_pq_copy)


def test_common_params() -> None:
    par = copy_param("zmin")
    assert par.default == 0.0
    assert par.value == 0.0
    assert par.dtype == float

    set_param_default("zmin", 0.1)
    par = copy_param("zmin")
    assert par.default == 0.1
    assert par.value == 0.1
    assert par.dtype == float


def test_catalog_utils() -> None:
    CatalogConfigBase.apply("rubin")
    CatalogConfigBase.band_name_dict()["u"] = "LSST_obs_u"
    assert CatalogConfigBase.active_tag() == "rubin"
    assert CatalogConfigBase.active_class().tag == "rubin"

    CatalogConfigBase.apply("roman_plus_rubin")
    RomanPlusRubinCatalogConfig.band_name_dict()["u"] = "LSST_obs_u"
    assert CatalogConfigBase.active_tag() == "roman_plus_rubin"
    assert CatalogConfigBase.active_class().tag == "roman_plus_rubin"

    for tag in [
        "hsc",
        "roman_3band",
        "roman_7band",
        "roman_rubin",
        "com_cam",
        "com_cam_gaap",
        "com_cam_euclid",
        "roman_3band_rubin",
        "roman_7band_rubin",
    ]:
        CatalogConfigBase.apply(tag)
        assert CatalogConfigBase.active_class().band_name_dict()

    CatalogConfigBase.apply("dc2")
    set_param_default("redshift_col", "redshift")

    a_class = CatalogConfigBase.get_class(
        "RomanPlusRubinCatalogConfig", "rail.utils.catalog_utils"
    )
    CatalogConfigBase.apply_class(a_class.__name__)

    assert "rubin" in CatalogConfigBase.subclasses()

    assert "RomanPlusRubinCatalogConfig" in CatalogConfigBase.subclasses_by_class()



import numpy as np
import qp
from ceci.config import StageConfig
from numpy.typing import NDArray
from scipy.integrate import simpson
from scipy.optimize import minimize_scalar

from rail.core.common_params import SHARED_PARAMS


class PointEstimationMixin:
    config_options = dict(
        calculated_point_estimates=SHARED_PARAMS,
        recompute_point_estimates=SHARED_PARAMS,
    )

    @property
    def config(self) -> StageConfig:
        raise NotImplementedError()

    def calculate_point_estimates(
        self, qp_dist: qp.Ensemble, grid: NDArray | list | None = None
    ) -> qp.Ensemble:
        """This function drives the calculation of point estimates for qp.Ensembles.
        It is defined here, and called from the `_process_chunk` method in the
        `CatEstimator` child classes.

        Parameters
        ----------
        qp_dist
            The qp Ensemble instance that contains posterior estimates.
        grid
            The grid on which to evaluate the point estimate. Note that not all
            point estimates require a grid to be provided, by default None.

        Returns
        -------
        qp.Ensemble
            The original `qp.Ensemble` with new ancillary point estimate data included.
            The `Ensemble.ancil` keys are ['mean', 'mode', 'median'].

        Notes
        -----
        If there are particularly efficient ways to calculate point estimates for
        a given `CatEstimator` subclass, the subclass can implement any of the
        `_calculate_<foo>_point_estimate` for any of the point estimate calculator
        methods, i.e.:

            - `_calculate_mode_point_estimate`
            - `_calculate_mean_point_estimate`
            - `_calculate_median_point_estimate`
        """

        ancil_dict: dict[str, NDArray] = dict()
        calculated_point_estimates = []

        existing_ancil = qp_dist.ancil
        if existing_ancil and not self.config.recompute_point_estimates:
            skip_zmode = "zmode" in existing_ancil
            skip_zmean = "zmean" in existing_ancil
            skip_zmedian = "zmedian" in existing_ancil
            skip_zbest = "zbest" in existing_ancil
        else:
            skip_zmode = False
            skip_zmean = False
            skip_zmedian = False
            skip_zbest = False

        if "calculated_point_estimates" in self.config:
            calculated_point_estimates = self.config["calculated_point_estimates"]

        if "zmode" in calculated_point_estimates and not skip_zmode:
            mode_value = self._calculate_mode_point_estimate(qp_dist, grid)
            ancil_dict.update(zmode=mode_value)

        if "zmean" in calculated_point_estimates and not skip_zmean:
            mean_value = self._calculate_mean_point_estimate(qp_dist)
            ancil_dict.update(zmean=mean_value)

        if "zmedian" in calculated_point_estimates and not skip_zmedian:
            median_value = self._calculate_median_point_estimate(qp_dist)
            ancil_dict.update(zmedian=median_value)

        if "zbest" in calculated_point_estimates and not skip_zbest:
            best_value = self._calculate_best_point_estimate(qp_dist)
            ancil_dict.update(zbest=best_value)

        if calculated_point_estimates:
            if qp_dist.ancil is None:
                qp_dist.set_ancil(ancil_dict)
            else:
                qp_dist.add_to_ancil(ancil_dict)

        return qp_dist

    def _calculate_mode_point_estimate(
        self, qp_dist: qp.Ensemble, grid: NDArray | list | None = None
    ) -> NDArray:
        """Calculates and returns the mode values for a set of posterior estimates
        in a qp.Ensemble instance.

        Parameters
        ----------
        qp_dist
            The qp Ensemble instance that contains posterior estimates.
        grid :
            The grid on which to evaluate the `mode` point estimate, if a grid is
            not provided, a default will be created at run time using `zmin`, `zmax`,
            and `nzbins`, by default None

        Returns
        -------
        NDArray
            The mode value for each posterior in the qp.Ensemble

        Raises
        ------
        KeyError
            If `grid` is not provided, one will be created using stage_config
            `zmin`, `zmax`, and `nzbins` keys. If any of those keys are missing,
            we'll raise a KeyError.
        """
        if grid is None:
            for key in ["zmin", "zmax", "nzbins"]:
                if key not in self.config:  # pragma: no cover
                    raise KeyError(
                        f"Expected `{key}` to be defined in stage "
                        "configuration dictionary in order to caluclate mode."
                    )

            grid = np.linspace(self.config.zmin, self.config.zmax, self.config.nzbins)

        return qp_dist.mode(grid=grid)

    def _calculate_mean_point_estimate(self, qp_dist: qp.Ensemble) -> NDArray:
        """Calculates and returns the mean values for a set of posterior estimates
        in a qp.Ensemble instance.

        Parameters
        ----------
        qp_dist
            The qp Ensemble instance that contains posterior estimates.

        Returns
        -------
        NDArray
            The mean value for each posterior in the qp.Ensemble
        """
        return qp_dist.mean()

    def _calculate_median_point_estimate(self, qp_dist: qp.Ensemble) -> NDArray:
        """Calculates and returns the median values for a set of posterior estimates
        in a qp.Ensemble instance.

        Parameters
        ----------
        qp_dist
            The qp Ensemble instance that contains posterior estimates.

        Returns
        -------
        NDArray
            The median value for each posterior in the qp.Ensemble
        """
        return qp_dist.median()

    def _calculate_best_point_estimate(
        self, qp_dist: qp.Ensemble, grid: NDArray | list | None = None
    ) -> NDArray:
        """
        Compute zx for all objects that minimizes ∫ dz P(z) * (zx - z)/(1 + z)

        Parameters
        ----------
        pdf_vals : np.ndarray
            Array of shape (N, k), PDF values for N objects evaluated on z_grid
        z_grid : np.ndarray
            1D array of redshift grid of shape (k,)

        Returns
        -------
        zx_array : np.ndarray
            Array of optimal zx values of shape (N,)
        """

        if grid is None:
            for key in ["zmin", "zmax", "nzbins"]:
                if key not in self.config:  # pragma: no cover
                    raise KeyError(
                        f"Expected `{key}` to be defined in stage "
                        "configuration dictionary in order to caluclate mode."
                    )

            grid = np.linspace(self.config.zmin, self.config.zmax, self.config.nzbins)
        elif isinstance(grid, list):  # pragma: no cover
            grid = np.array(grid)

        assert isinstance(grid, np.ndarray)

        pdf_vals = qp_dist.pdf(grid)
        N = pdf_vals.shape[0]
        zx_array = np.zeros(N)

        for i in range(N):
            pz = pdf_vals[i]

            def risk(zx: NDArray) -> float:
                integrand = pz * loss(zx, grid)
                return simpson(integrand, grid)

            def loss(zx: NDArray, grid: NDArray, gamma: float=0.15) -> np.ndarray:
                dz = (zx - grid) / (1 + grid)
                return 1 - 1 / (1 + (dz / gamma) ** 2)

            result = minimize_scalar(risk, bounds=(grid[0], grid[-1]), method="bounded")
            zx_array[i] = result.x

        return zx_array



import os
from typing import Any

import qp

from rail.core.data import QPHandle, TableHandle
from rail.core.stage import RailStage
from rail.estimation.algos import naive_stack, point_est_hist, var_inf
from rail.utils.path_utils import RAILDIR

testdata = os.path.join(RAILDIR, "rail/examples_data/testdata/output_BPZ_lite.hdf5")
tomobins = os.path.join(RAILDIR, "rail/examples_data/testdata/output_tomo.hdf5")
DS = RailStage.data_store


def one_algo(
    key: str, summarizer_class: type[RailStage], summary_kwargs: dict[str, Any]
) -> qp.Ensemble:
    """
    A basic test of running an summaizer subclass
    Run summarize
    """
    DS.__class__.allow_overwrite = True
    DS.clear()
    test_data = DS.read_file("test_data", QPHandle, testdata)
    summarizer = summarizer_class.make_stage(name=key, **summary_kwargs)
    summary_ens = summarizer.summarize(test_data)
    os.remove(
        summarizer.get_output(summarizer.get_aliased_tag("output"), final_name=True)
    )
    os.remove(
        summarizer.get_output(summarizer.get_aliased_tag("single_NZ"), final_name=True)
    )
    return summary_ens


def one_mask_algo(
    key: str, summarizer_class: type[RailStage], summary_kwargs: dict[str, Any]
) -> list[qp.Ensemble]:
    """
    A basic test of running an summaizer subclass
    Run summarize
    """
    DS.__class__.allow_overwrite = True
    DS.clear()
    test_data = DS.read_file("test_data", QPHandle, testdata)
    tomo_bins = DS.read_file("tomo_bins", TableHandle, tomobins)

    summarizer = summarizer_class.make_stage(name=key, selected_bin=1, **summary_kwargs)
    summary_ens = summarizer.summarize(test_data, tomo_bins)
    os.remove(
        summarizer.get_output(summarizer.get_aliased_tag("output"), final_name=True)
    )
    os.remove(
        summarizer.get_output(summarizer.get_aliased_tag("single_NZ"), final_name=True)
    )

    summarizer_2 = summarizer_class.make_stage(name=f"{key}_2", **summary_kwargs)
    summary_2_ens = summarizer_2.summarize(test_data, None)
    os.remove(
        summarizer_2.get_output(summarizer_2.get_aliased_tag("output"), final_name=True)
    )
    os.remove(
        summarizer_2.get_output(
            summarizer_2.get_aliased_tag("single_NZ"), final_name=True
        )
    )

    return [summary_ens, summary_2_ens]


def test_naive_stack() -> None:
    """Basic end to end test for the Naive stack informer to estimator stages"""
    naive_stack_informer_stage = naive_stack.NaiveStackInformer.make_stage()
    naive_stack_informer_stage.inform()

    summary_config_dict: dict = {}
    summarizer_class = naive_stack.NaiveStackSummarizer
    _ = one_algo("NaiveStack", summarizer_class, summary_config_dict)
    os.remove(
        naive_stack_informer_stage.get_output(
            naive_stack_informer_stage.get_aliased_tag("model"), final_name=True
        )
    )


def test_point_estimate_hist() -> None:
    """Basic end to end test for the point estimate histogram informer to estimator
    stages
    """
    point_est_informer_stage = point_est_hist.PointEstHistInformer.make_stage()
    point_est_informer_stage.inform()

    summary_config_dict: dict = {}
    summarizer_class = point_est_hist.PointEstHistSummarizer
    _ = one_algo("PointEstimateHist", summarizer_class, summary_config_dict)
    os.remove(
        point_est_informer_stage.get_output(
            point_est_informer_stage.get_aliased_tag("model"), final_name=True
        )
    )


def test_var_inference_stack() -> None:
    """Basic end to end test for the var inference informer to estimator stages"""
    var_inf_informer_stage = var_inf.VarInfStackInformer.make_stage()
    var_inf_informer_stage.inform()

    summary_config_dict: dict = {}
    summarizer_class = var_inf.VarInfStackSummarizer
    _ = one_algo("VariationalInference", summarizer_class, summary_config_dict)
    os.remove(
        var_inf_informer_stage.get_output(
            var_inf_informer_stage.get_aliased_tag("model"), final_name=True
        )
    )


def test_naive_stack_masked() -> None:
    """Basic end to end test for the Naive stack informer to estimator stages"""
    summary_config_dict = dict(
        chunk_size=5,
    )
    summarizer_class = naive_stack.NaiveStackMaskedSummarizer
    _ = one_mask_algo("NaiveStack", summarizer_class, summary_config_dict)
    _ = one_algo("NaiveStack", summarizer_class, summary_config_dict)


def test_point_estimate_hist_masked() -> None:
    """Basic end to end test for the point estimate histogram informer to estimator
    stages
    """
    summary_config_dict = dict(
        chunk_size=5,
    )
    summarizer_class = point_est_hist.PointEstHistMaskedSummarizer
    _ = one_mask_algo("PointEstimateHist", summarizer_class, summary_config_dict)
    _ = one_algo("PointEstimateHist", summarizer_class, summary_config_dict)



import os
from typing import Any

import yaml

import rail.stages
from rail.cli.rail.options import GitMode
from rail.core import RailEnv
from rail.core.stage import RailPipeline
from rail.utils import catalog_utils
from rail.utils.path_utils import RAILDIR


def render_nb(
    outdir: str,
    clear_output: bool,
    dry_run: bool,
    inputs: list[str],
    skip: list[str],
    **_kwargs: Any,
) -> None:
    command = "jupyter nbconvert"
    options = "--to html"

    status = {}

    for nb_file in inputs:
        if nb_file in skip:
            continue
        subdir = os.path.dirname(nb_file).split("/")[-1]
        basename = os.path.splitext(os.path.basename(nb_file))[0]
        outfile = os.path.join("..", "..", outdir, f"{subdir}/{basename}.html")
        relpath = os.path.join(outdir, f"{subdir}")

        try:
            print(relpath)
            os.makedirs(relpath)
        except FileExistsError:
            pass

        if clear_output:
            comline = f"{command} --clear-output {nb_file}"
        else:
            comline = f"{command} {options} --output {outfile} --execute {nb_file}"

        if dry_run:
            render = 0
            print(comline)
        else:
            render = os.system(comline)
        status[nb_file] = render

    failed_notebooks = []
    for key, val in status.items():
        print(f"{key} {val}")
        if val != 0:  # pragma: no cover
            failed_notebooks.append(key)

    if failed_notebooks:  # pragma: no cover
        raise ValueError(f"The following notebooks failed {str(failed_notebooks)}")


def clone_source(
    outdir: str, git_mode: GitMode, dry_run: bool, package_file: str
) -> None:  # pragma: no cover
    with open(package_file, encoding="utf-8") as pfile:
        package_dict = yaml.safe_load(pfile)

    for key, _val in package_dict.items():
        if os.path.exists(f"{outdir}/{key}"):
            print(f"Skipping existing {outdir}/{key}")
            continue

        if git_mode == GitMode.ssh:
            com_line = f"git clone https://github.com/LSSTDESC/{key}.git {outdir}/{key}"
        elif git_mode == GitMode.https:
            com_line = f"git clone git@github.com:LSSTDESC/{key}.git {outdir}/{key}"
        elif git_mode == GitMode.cli:
            com_line = f"gh repo clone LSSTDESC/{key} {outdir}/{key}"

        if dry_run:
            print(com_line)
        else:
            os.system(com_line)


def update_source(outdir: str, dry_run: bool, package_file: str) -> None:
    with open(package_file, encoding="utf-8") as pfile:
        package_dict = yaml.safe_load(pfile)

    currentpath = os.path.abspath(".")
    for key, _val in package_dict.items():
        abspath = os.path.abspath(f"{outdir}/{key}")

        if os.path.exists(f"{outdir}/{key}") is not True:  # pragma: no cover
            print(f"Package {outdir}/{key} does not exist!")
            continue

        com_line = f"cd {abspath} && git pull && cd {currentpath}"

        if dry_run:
            print(com_line)
        else:  # pragma: no cover
            os.system(com_line)


def git_status(outdir: str, dry_run: bool, package_file: str) -> None:
    with open(package_file, encoding="utf-8") as pfile:
        package_dict = yaml.safe_load(pfile)

    logfile = os.path.abspath("./git_status.log")

    try:
        os.unlink(logfile)
    except:
        pass
    os.system(f"touch {logfile}")
    
    currentpath = os.path.abspath(".")
    for key, _val in package_dict.items():
        abspath = os.path.abspath(f"{outdir}/{key}")

        if os.path.exists(f"{outdir}/{key}") is not True:  # pragma: no cover
            print(f"Package {outdir}/{key} does not exist!")
            continue

        os.system(f"echo '------- {key} ---------' >> {logfile}")
        com_line = f"cd {abspath} && git status >> {logfile} && cd {currentpath}"

        if dry_run:
            print(com_line)
        else:  # pragma: no cover
            os.system(com_line)

    print(f"Wrote output to {logfile}")


def git_diff(outdir: str, dry_run: bool, package_file: str) -> None:
    with open(package_file, encoding="utf-8") as pfile:
        package_dict = yaml.safe_load(pfile)

    logfile = os.path.abspath("./git_diff.log")

    try:
        os.unlink(logfile)
    except:
        pass
    os.system(f"touch {logfile}")
    
    currentpath = os.path.abspath(".")
    for key, _val in package_dict.items():
        abspath = os.path.abspath(f"{outdir}/{key}")

        if os.path.exists(f"{outdir}/{key}") is not True:  # pragma: no cover
            print(f"Package {outdir}/{key} does not exist!")
            continue

        os.system(f"echo '------- {key} ---------' >> {logfile}")
        com_line = f"cd {abspath} && git diff >> {logfile} && cd {currentpath}"

        if dry_run:
            print(com_line)
        else:  # pragma: no cover
            os.system(com_line)

    print(f"Wrote output to {logfile}")
    

def git_describe(outdir: str, dry_run: bool, package_file: str) -> None:
    with open(package_file, encoding="utf-8") as pfile:
        package_dict = yaml.safe_load(pfile)

    logfile = os.path.abspath("./git_tags.log")

    try:
        os.unlink(logfile)
    except:
        pass    
    os.system(f"touch {logfile}")
    
    currentpath = os.path.abspath(".")
    for key, _val in package_dict.items():
        abspath = os.path.abspath(f"{outdir}/{key}")

        if os.path.exists(f"{outdir}/{key}") is not True:  # pragma: no cover
            print(f"Package {outdir}/{key} does not exist!")
            continue

        os.system(f"echo '{key}' >> {logfile}")
        com_line = f"cd {abspath} && git describe --tags >> {logfile} && cd {currentpath}"

        if dry_run:
            print(com_line)
        else:  # pragma: no cover
            os.system(com_line)

    print(f"Wrote output to {logfile}")
    
    

def install(outdir: str, from_source: bool, dry_run: bool, package_file: str) -> None:
    with open(package_file, encoding="utf-8") as pfile:
        package_dict = yaml.safe_load(pfile)

    for key, val in package_dict.items():
        if not from_source:
            com_line = f"pip install {val}"
        else:
            if not os.path.exists(f"{outdir}/{key}"):  # pragma: no cover
                print(f"Skipping missing {outdir}/{key}")
                continue
            com_line = f"pip install -e {outdir}/{key}"

        if dry_run:
            print(com_line)
        else:  # pragma: no cover
            os.system(com_line)


def info(**kwargs: Any) -> None:
    rail.stages.import_and_attach_all()

    print_all = kwargs.get("print_all", False)
    if kwargs.get("print_packages") or print_all:
        print("======= Printing RAIL packages ==============")
        RailEnv.print_rail_packages()
        print("\n\n")
    if kwargs.get("print_namespaces") or print_all:
        print("======= Printing RAIL namespaces ==============")
        RailEnv.print_rail_namespaces()
        print("\n\n")
    if kwargs.get("print_modules") or print_all:
        print("======= Printing RAIL modules ==============")
        RailEnv.print_rail_modules()
        print("\n\n")
    if kwargs.get("print_tree") or print_all:
        print("======= Printing RAIL source tree ==============")
        RailEnv.print_rail_namespace_tree()
        print("\n\n")
    if kwargs.get("print_stages") or print_all:
        print("======= Printing RAIL stages ==============")
        RailEnv.print_rail_stage_dict()
        print("\n\n")


def get_data(verbose: bool, **kwargs: Any) -> None:  # pragma: no cover
    standard_data_files = [
        {
            "local_path": "rail/examples_data/goldenspike_data/data/base_catalog.pq",
            "remote_path": "https://portal.nersc.gov/cfs/lsst/PZ/base_catalog.pq",
        }
    ]
    bpz_data_files = [
        {
            "local_path": "rail/examples_data/estimation_data/data/nonphysical_dc2_templates.tar",
            "remote_path": "https://portal.nersc.gov/cfs/lsst/PZ/nonphysical_dc2_templates.tar",
        },
        {
            "local_path": "rail/examples_data/estimation_data/data/test_dc2_training_9816_broadtypes.hdf5",
            "remote_path": "https://portal.nersc.gov/cfs/lsst/PZ/test_dc2_training_9816_broadtypes.hdf5",
        },
        {
            "local_path": "rail/examples_data/estimation_data/data/test_dc2_train_customtemp_broadttypes.hdf5",  # pylint: disable=line-too-long
            "remote_path": "https://portal.nersc.gov/cfs/lsst/PZ/test_dc2_train_customtemp_broadttypes.hdf5",
        },
    ]

    data_files = standard_data_files
    if kwargs.get("bpz_demo_data"):
        # The bpz demo data is quarantined into its own flag, as it contains some
        # non-physical features that would add systematics if run on any real data.
        # This data should NOT be used for any science with real data!
        data_files = bpz_data_files
        print("Downloading BPZ demo data...")
        print(
            "(Note: you can run get-data without the bpz-demo-data flag to download standard data.)"
        )

    for data_file in data_files:
        local_abs_path = os.path.join(RAILDIR, data_file["local_path"])
        if verbose:
            print(
                f"Check file exists: {local_abs_path} ({os.path.exists(local_abs_path)})"
            )
        if not os.path.exists(local_abs_path):
            os.system(
                f'curl -o {local_abs_path} {data_file["remote_path"]} --create-dirs'
            )


def build_pipeline(
    pipeline_class: str,
    output_yaml: str,
    catalog_tag: str | None = None,
    input_dict: dict | None = None,
    stages_config: dict | None = None,
    output_dir: str = ".",
    log_dir: str | None = None,
    **kwargs: Any,
) -> None:
    tokens = pipeline_class.split(".")
    module = ".".join(tokens[:-1])
    class_name = tokens[-1]

    if catalog_tag:
        catalog_utils.apply_defaults(catalog_tag)

    if log_dir is None:
        log_dir = os.path.join(output_dir, "logs", class_name)

    __import__(module)
    RailPipeline.build_and_write(
        class_name,
        output_yaml,
        input_dict,
        stages_config,
        output_dir,
        log_dir,
        **kwargs,
    )



"""
Abstract base classes defining classifiers.
"""

import gc
from typing import Any

import qp

from rail.core.common_params import SHARED_PARAMS
from rail.core.data import (DataHandle, Hdf5Handle, ModelHandle, ModelLike,
                            QPHandle, TableHandle, TableLike)
from rail.core.stage import RailStage


class CatClassifier(RailStage):  # pragma: no cover
    """The base class for assigning classes to catalogue-like table.

    Classifier uses a generic "model", the details of which depends on the sub-class.

    CatClassifier take as "input" a catalogue-like table, assign each object into
    a tomographic bin, and provide as "output" a tabular data which can be appended
    to the catalogue.
    """

    name = "CatClassifier"
    config_options = RailStage.config_options.copy()
    config_options.update(
        chunk_size=SHARED_PARAMS,
        hdf5_groupname=SHARED_PARAMS,
    )
    inputs = [("model", ModelHandle), ("input", TableHandle)]
    outputs = [("output", TableHandle)]

    def __init__(self, args: Any, **kwargs: Any) -> None:
        """Initialize Classifier"""
        super().__init__(args, **kwargs)
        self._output_handle: TableHandle | None = None
        self.model: ModelLike | None = None
        if not isinstance(args, dict):  # pragma: no cover
            args = vars(args)
        self.open_model(**args)

    def classify(self, input_data: TableLike) -> DataHandle:
        """The main run method for the classifier, should be implemented
        in the specific subclass.

        This will attach the input_data to this `CatClassifier`
        (for introspection and provenance tracking).

        Then it will call the run() and finalize() methods, which need to
        be implemented by the sub-classes.

        The run() method will need to register the data that it creates to this Classifier
        by using `self.add_data('output', output_data)`.

        Finally, this will return a TableHandle providing access to that output data.

        Parameters
        ----------
        input_data
            A dictionary of all input data

        Returns
        -------
        DataHandle
            Class assignment for each galaxy.
        """
        self.set_data("input", input_data)
        self.run()
        self.finalize()
        return self.get_handle("output")


class PZClassifier(RailStage):
    """The base class for assigning classes (tomographic bins) to per-galaxy PZ
    estimates.

    PZClassifier takes as "input" a `qp.Ensemble` with per-galaxy PDFs, and
    provides as "output" tabular data which can be appended to the catalogue.
    """

    name = "PZClassifier"
    config_options = RailStage.config_options.copy()
    config_options.update(chunk_size=SHARED_PARAMS)
    inputs = [("input", QPHandle)]
    outputs = [("output", Hdf5Handle)]

    def __init__(self, args: Any, **kwargs: Any) -> None:
        """Initialize the PZClassifier."""
        super().__init__(args, **kwargs)
        self._output_handle: DataHandle | None = None

    def classify(self, input_data: qp.Ensemble) -> DataHandle:
        """The main run method for the classifier, should be implemented
        in the specific subclass.

        This will attach the input_data to this `PZClassifier`
        (for introspection and provenance tracking).

        Then it will call the run() and finalize() methods, which need to
        be implemented by the sub-classes.

        The run() method will need to register the data that it creates to this
        Classifier by using `self.add_data('output', output_data)`.

        The run() method relies on the _process_chunk() method, which should be
        implemented by subclasses to perform the actual classification on each
        chunk of data. The results from each chunk are then combined in the
        _finalize_run() method. (Alternatively, override run() in a subclass to
        perform the classification without parallelization.)

        Finally, this will return a TableHandle providing access to that output data.

        Parameters
        ----------
        input_data
            Per-galaxy p(z), and any ancilary data associated with it

        Returns
        -------
        DataHandle
            Class assignment for each galaxy, typically in the form of a
            dictionary with IDs and class labels.
        """
        self.set_data("input", input_data)
        self.run()
        self.finalize()
        return self.get_handle("output")

    def _finalize_run(self) -> None:
        """Finalize the classification process after processing all chunks."""
        assert self._output_handle is not None
        self._output_handle.finalize_write()

    def _process_chunk(
        self, start: int, end: int, data: qp.Ensemble, first: bool
    ) -> None:
        """Process a chunk of data.

        This method should be implemented in subclasses to perform the actual
        classification on a chunk of data.

        Parameters
        ----------
        start
            The starting index of the chunk.

        end
            The ending index of the chunk.

        data
            The data chunk to be processed.

        first
            True if this is the first chunk, False otherwise.
        """
        raise NotImplementedError(
            f"{self.name}._process_chunk is not implemented"
        )  # pragma: no cover

    def _do_chunk_output(
        self, class_id: TableLike, start: int, end: int, first: bool
    ) -> None:
        """Handle the output of a processed chunk.

        Parameters
        ----------
        class_id
            The classification results for the chunk.

        start
            The starting index of the chunk.

        end
            The ending index of the chunk.

        first
            True if this is the first chunk, False otherwise.
        """
        if first:
            self._output_handle = self.add_handle("output", data=class_id)
            assert self._output_handle is not None
            self._output_handle.initialize_write(
                self._input_length, communicator=self.comm
            )
        assert self._output_handle is not None
        self._output_handle.set_data(class_id, partial=True)
        self._output_handle.write_chunk(start, end)

    def run(self) -> None:
        """Processes the input data in chunks and performs classification.

        This method iterates over chunks of the input data, calling the
        _process_chunk method for each chunk to perform the actual classification.

        The _process_chunk method should be implemented by subclasses to define
        the specific classification logic.
        """
        test_data = self.get_data("input")

        iterator = self.input_iterator("input")
        first = True

        for start, end, test_data in iterator:
            # print(f"Process {self.rank} running estimator on chunk {start} - {end}")
            self._process_chunk(start, end, test_data, first)
            first = False
            # Running garbage collection manually seems to be needed
            # to avoid memory growth for some estimators
            gc.collect()
        self._finalize_run()



import os
import pickle
from typing import Any

import numpy as np
import pytest

from rail.core.model import Model
from rail.estimation.algos.train_z import trainZmodel


def pickle_dump(obj: Any, path: str) -> None:
    with open(path, "wb") as fout:
        pickle.dump(obj=obj, file=fout, protocol=pickle.HIGHEST_PROTOCOL)


def test_model() -> None:
    array_data = np.array([None])
    dict_data = dict(a=5, b=6)
    train_z_data = trainZmodel(np.linspace(0.0, 3.0, 301), np.ones(301), 1.5)
    model = Model(array_data, "dummy", 0, dict(alice="bob"))
    _model_2 = Model(array_data, "dummy", 0)

    pickle_dump(array_data, "array_raw.pickle")
    _read_array_data = Model.read("array_raw.pickle", "dummy", 0, dict(alice="bob"))
    _read_array_data_2 = Model.read("array_raw.pickle", "dummy", 0)

    Model.wrap("array_raw.pickle", "array_wrap.pickle", "dummy", 0, dict(alice="bob"))

    Model.dump(array_data, "array_data.pickle", "dummy", 0)
    Model.dump(dict_data, "dict_data.pickle", "dummy", 0, dict(alice="bob"))
    Model.dump(train_z_data, "train_z_data.pickle", "dummy", 0, dict(alice="bob"))
    Model.dump(model, "model_data.pickle")

    check_array_data = Model.read("array_data.pickle")
    _check_dict_data = Model.read("dict_data.pickle")
    _check_train_z_data = Model.read("train_z_data.pickle")
    _check_model_data = Model.read("model_data.pickle")

    Model.validate(check_array_data, "dummy", 0)
    with pytest.raises(TypeError):
        Model.validate(check_array_data, "dummies", 0)

    with pytest.raises(TypeError):
        Model.validate(check_array_data, "dummy", 1)

    os.remove("array_data.pickle")
    os.remove("array_raw.pickle")
    os.remove("array_wrap.pickle")
    os.remove("dict_data.pickle")
    os.remove("model_data.pickle")
    os.remove("train_z_data.pickle")



"""
A classifier that uses pz point estimate to assign
tomographic bins with uniform binning. 
"""

import numpy as np
import qp
from ceci.config import StageParameter as Param

from rail.core.data import Hdf5Handle
from rail.estimation.classifier import PZClassifier


class UniformBinningClassifier(PZClassifier):
    """Classifier that simply assigns tomographic bins based on a point estimate
    according to SRD.
    """

    name = "UniformBinningClassifier"
    config_options = PZClassifier.config_options.copy()
    config_options.update(
        id_name=Param(
            str,
            "",
            msg="Column name for the object ID in the input data, if empty the row index is used as the ID.",
        ),
        point_estimate=Param(str, "zmode", msg="Which point estimate to use"),
        zbin_edges=Param(
            list,
            [],
            msg="The tomographic redshift bin edges."
            "If this is given (contains two or more entries), all settings below will be ignored.",
        ),
        zmin=Param(float, 0.0, msg="Minimum redshift of the sample"),
        zmax=Param(float, 3.0, msg="Maximum redshift of the sample"),
        nbins=Param(int, 5, msg="Number of tomographic bins"),
        no_assign=Param(int, -99, msg="Value for no assignment flag"),
    )
    outputs = [("output", Hdf5Handle)]

    def _process_chunk(
        self, start: int, end: int, data: qp.Ensemble, first: bool
    ) -> None:
        """Process a chunk of data for uniform binning classification.

        Parameters
        ----------
        start
            The starting index of the chunk

        end
            The ending index of the chunk

        data
            The data chunk to be processed

        first
            True if this is the first chunk, False otherwise.
        """
        try:
            zb = data.ancil[self.config.point_estimate]
        except KeyError as missing_key:
            raise KeyError(
                f"{self.config.point_estimate} is not contained "
                "in the data ancil, you will need to compute it explicitly."
            ) from missing_key

        try:
            _npdf = data.npdf
        except KeyError as missing_key:
            raise KeyError(
                f"npdf is not a supported attribute of {type(data)}. "
                "Are you sure you don't mean to be using a qp ensemble?"
            ) from missing_key

        # binning options
        if len(self.config.zbin_edges) >= 2:
            # this overwrites all other key words
            # linear binning defined by zmin, zmax, and nbins
            bin_index = np.digitize(zb, self.config.zbin_edges)
            # assign -99 to objects not in any bin:
            bin_index[bin_index == 0] = self.config.no_assign
            bin_index[bin_index == len(self.config.zbin_edges)] = self.config.no_assign
        else:
            # linear binning defined by zmin, zmax, and nbins
            bin_index = np.digitize(
                zb,
                np.linspace(self.config.zmin, self.config.zmax, self.config.nbins + 1),
            )
            # assign -99 to objects not in any bin:
            bin_index[bin_index == 0] = self.config.no_assign
            bin_index[bin_index == (self.config.nbins + 1)] = self.config.no_assign

        if self.config.id_name != "":
            # below is commented out and replaced by a redundant line
            # because the data doesn't have ID yet
            # obj_id = data[self.config.id_name]
            obj_id = np.arange(data.npdf)
        elif self.config.id_name == "":
            # ID set to row index
            obj_id = np.arange(data.npdf)
            self.config.id_name = "row_index"

        class_id = {self.config.id_name: obj_id, "class_id": bin_index}
        self._do_chunk_output(class_id, start, end, first)



from typing import Any

import qp


class MetricEvaluator:
    """A superclass for metrics evaluations"""

    def __init__(self, qp_ens: qp.Ensemble) -> None:
        """Class constructor.
        Parameters
        ----------
        qp_ens: qp.Ensemble object
            PDFs as qp.Ensemble
        """
        self._qp_ens = qp_ens

    def evaluate(self) -> Any:  # pragma: no cover
        """
        Evaluates the metric a function of the truth and prediction

        Returns
        -------
        metric: dictionary
            value of the metric and statistics thereof
        """
        raise NotImplementedError



from typing import Any

import numpy as np
from ceci.config import StageParameter as Param
from qp.metrics.concrete_metric_classes import DistToPointMetric

from rail.core.data import QPHandle, TableHandle
from rail.evaluation.evaluator import Evaluator


class DistToPointEvaluator(Evaluator):
    """Evaluate the performance of a photo-z estimator against reference point estimate"""

    name = "DistToPointEvaluator"
    config_options = Evaluator.config_options.copy()
    config_options.update(
        limits=Param(
            list,
            [0.0, 3.0],
            required=False,
            msg="The default end points for calculating metrics on a grid.",
        ),
        dx=Param(
            float,
            0.01,
            required=False,
            msg="The default step size when calculating metrics on a grid.",
        ),
        quantile_grid=Param(
            list,
            np.linspace(0, 1, 100).tolist(),
            required=False,
            msg="The quantile value grid on which to evaluate the CDF values. (0, 1)",
        ),
        x_grid=Param(
            list,
            np.linspace(0, 2.5, 301).tolist(),
            required=False,
            msg="The x-value grid at which to evaluate the pdf values.",
        ),
        hdf5_groupname=Param(
            str, "photometry", required=False, msg="HDF5 Groupname for truth table."
        ),
        reference_dictionary_key=Param(
            str,
            "redshift",
            required=False,
            msg="The key in the `truth` dictionary where the redshift data is stored.",
        ),
    )
    inputs = [("input", QPHandle), ("truth", TableHandle)]

    metric_base_class = DistToPointMetric

    def _process_chunk(self, data_tuple: Any, first: bool) -> None:
        start = data_tuple[0]
        end = data_tuple[1]
        estimate_data = data_tuple[2]
        reference_data = data_tuple[3][self.config.reference_dictionary_key]

        self._process_all_chunk_metrics(
            estimate_data, reference_data, start, end, first
        )

    def _process_all(self, data_tuple: Any) -> None:
        estimate_data = data_tuple[0]
        if self.config.hdf5_groupname:
            reference_data = data_tuple[1][self.config.hdf5_groupname][
                self.config.reference_dictionary_key
            ]
        else:  # pragma: no cover
            reference_data = data_tuple[1][self.config.reference_dictionary_key]
        self._process_all_metrics(estimate_data, reference_data)



import rail
from rail.core import RailEnv
from rail.core.stage import RailStage
from rail.creation.degrader import Degrader
from rail.creation.degraders.addRandom import AddColumnOfRandom
from rail.creation.degraders.quantityCut import QuantityCut
from rail.creation.engine import Creator, Modeler, PosteriorCalculator
from rail.creation.noisifier import Noisifier
from rail.creation.selector import Selector
from rail.estimation.algos.equal_count import EqualCountClassifier
from rail.estimation.algos.naive_stack import (NaiveStackInformer,
                                               NaiveStackSummarizer)
from rail.estimation.algos.point_est_hist import (PointEstHistInformer,
                                                  PointEstHistSummarizer)
from rail.estimation.algos.random_gauss import (RandomGaussEstimator,
                                                RandomGaussInformer)
from rail.estimation.algos.train_z import TrainZEstimator, TrainZInformer
from rail.estimation.algos.true_nz import TrueNZHistogrammer
from rail.estimation.algos.uniform_binning import UniformBinningClassifier
from rail.estimation.algos.var_inf import (VarInfStackInformer,
                                           VarInfStackSummarizer)
from rail.estimation.classifier import CatClassifier, PZClassifier
from rail.estimation.estimator import CatEstimator, PzEstimator
from rail.estimation.informer import CatInformer, PzInformer

from rail.estimation.summarizer import (CatSummarizer, PZSummarizer,
                                        SZPZSummarizer)
from rail.evaluation.dist_to_dist_evaluator import DistToDistEvaluator
from rail.evaluation.dist_to_point_evaluator import DistToPointEvaluator
from rail.evaluation.evaluator import Evaluator, OldEvaluator
from rail.evaluation.point_to_point_evaluator import PointToPointEvaluator
from rail.evaluation.single_evaluator import SingleEvaluator
from rail.tools.table_tools import ColumnMapper, RowSelector, TableConverter

__all__ = [
    "CatEstimator",
    "CatClassifier",
    "PZClassifier",
    "CatSummarizer",
    "PZSummarizer",
    "SZPZSummarizer",
    "NaiveStackInformer",
    "NaiveStackSummarizer",
    "RandomGaussInformer",
    "RandomGaussEstimator",
    "PointEstHistInformer",
    "PointEstHistSummarizer",
    "TrainZInformer",
    "TrainZEstimator",
    "VarInfStackInformer",
    "VarInfStackSummarizer",
    "UniformBinningClassifier",
    "EqualCountClassifier",
    "TrueNZHistogrammer",
    "Degrader",
    "AddColumnOfRandom",
    "QuantityCut",
    "Modeler",
    "Creator",
    "PosteriorCalculator",
    "Noisifier",
    "Selector",
    "DistToDistEvaluator",
    "DistToPointEvaluator",
    "Evaluator",
    "OldEvaluator",
    "PointToPointEvaluator",
    "SingleEvaluator",
    "ColumnMapper",
    "RowSelector",
    "TableConverter",
]


def import_and_attach_all() -> None:
    """Import all the packages in the rail ecosystem and attach them to this module"""
    RailEnv.import_all_packages()
    RailEnv.attach_stages(rail.stages)
    for xx in RailStage.pipeline_stages:
        rail.stages.__all__.append(xx)



import pytest

from rail.utils.testing_utils import build_and_read_pipeline


@pytest.mark.parametrize(
    "pipeline_class",
    [
        "rail.pipelines.estimation.train_z_pipeline.TrainZPipeline",
    ],
)
def test_build_and_read_pipeline(pipeline_class: str) -> None:
    build_and_read_pipeline(pipeline_class)



# This is a deprecative path for importing RAILDIR, it is only for the transition period
# to move RAILDIR and find_rail_file to utils.path_utils

from rail.utils.path_utils import RAILDIR, find_rail_file

__all__ = [
    "find_rail_file",
    "RAILDIR",
]



"""
A summarizer that simple makes a histogram of a point estimate
"""

from typing import Any, Iterable

import numpy as np
import qp
from ceci.config import StageParameter as Param
from scipy.special import digamma
from scipy.stats import dirichlet

from rail.core.data import QPHandle
from rail.estimation.informer import PzInformer
from rail.estimation.summarizer import PZSummarizer

TEENY = 1.0e-15


class VarInfStackInformer(PzInformer):
    """Placeholder Informer"""

    name = "VarInfStackInformer"
    config_options = PzInformer.config_options.copy()

    def _finalize_run(self) -> None:
        self.model = np.array([None])
        PzInformer._finalize_run(self)


class VarInfStackSummarizer(PZSummarizer):
    """Variational inference summarizer based on notebook created by Markus Rau
    The summzarizer is appropriate for the likelihoods returned by
    template-based codes, for which the NaiveSummarizer are not appropriate.

    """

    name = "VarInfStackSummarizer"
    config_options = PZSummarizer.config_options.copy()
    config_options.update(
        zmin=Param(float, 0.0, msg="The minimum redshift of the z grid"),
        zmax=Param(float, 3.0, msg="The maximum redshift of the z grid"),
        nzbins=Param(int, 301, msg="The number of gridpoints in the z grid"),
        seed=Param(int, 87, msg="random seed"),
        niter=Param(
            int, 100, msg="The number of iterations in the variational inference"
        ),
        nsamples=Param(
            int, 500, msg="The number of samples used in dirichlet uncertainty"
        ),
    )
    inputs = [("input", QPHandle)]
    outputs = [("output", QPHandle), ("single_NZ", QPHandle)]

    def __init__(self, args: Any, **kwargs: Any) -> None:
        super().__init__(args, **kwargs)
        self.zgrid: np.ndarray | None = None

    def _setup_iterator(self) -> Iterable:
        input_handle = self.get_handle("input", allow_missing=True)
        try:
            self.config.hdf5_groupname
        except Exception:
            self.config.hdf5_groupname = None
        input_length = input_handle.size(groupname=self.config.hdf5_groupname)
        chunk_size = int(np.ceil(input_length / self.size))

        iterator = self.input_iterator("input", chunk_size=chunk_size)
        return iterator

    def run(self) -> None:
        # Redefining the chunk size so that all of the data is distributed at once in the
        # nodes. This would fill all the memory if not enough nodes are allocated

        iterator = self._setup_iterator()
        self.zgrid = np.linspace(self.config.zmin, self.config.zmax, self.config.nzbins)
        first = True
        for s, e, test_data in iterator:
            print(f"Process {self.rank} running estimator on chunk {s:,} - {e:,}")
            alpha_trace = self._process_chunk(s, e, test_data, first)
            first = False

        if self.rank == 0:
            # old way of just spitting out a single distribution
            # qp_d = qp.Ensemble(qp.interp, data=dict(xvals=self.zgrid, yvals=alpha_trace))
            # instead, sample and save the samples
            rng = np.random.default_rng(seed=self.config.seed)
            sample_pz = dirichlet.rvs(
                alpha_trace, size=self.config.nsamples, random_state=rng
            )
            qp_d = qp.Ensemble(
                qp.interp, data=dict(xvals=self.zgrid, yvals=alpha_trace)
            )

            sample_ens = qp.Ensemble(
                qp.interp, data=dict(xvals=self.zgrid, yvals=sample_pz)
            )
            self.add_data("output", sample_ens)
            self.add_data("single_NZ", qp_d)

    def _process_chunk(
        self, _start: int, _end: int, test_data: qp.Ensemble, first: bool
    ) -> np.ndarray:
        if not first:  # pragma: no cover
            raise ValueError(
                "This algorithm needs all data in memory at once, increase nprocess or chunk size."
            )

        # Initiallizing arrays
        assert self.zgrid is not None
        alpha_trace = np.ones(len(self.zgrid))
        init_trace = np.ones(len(self.zgrid))
        pdf_vals = test_data.pdf(self.zgrid)
        log_pdf_vals = np.log(np.array(pdf_vals) + TEENY)
        for _ in range(self.config.niter):
            dig = np.array(
                [digamma(kk) - digamma(np.sum(alpha_trace)) for kk in alpha_trace]
            )
            matrix_grid = np.exp(dig + log_pdf_vals)
            gamma_matrix = np.array([kk / np.sum(kk) for kk in matrix_grid])
            for kk in matrix_grid:
                break
            nk_partial = np.sum(gamma_matrix, axis=0)
            if self.comm is not None:  # pragma: no cover
                nk = self.comm.allreduce(nk_partial)
            else:
                nk = nk_partial
            alpha_trace = nk + init_trace
        return alpha_trace



"""
Abstract base classes defining Summarizers of the redshift distribution of an ensemble of galaxies
"""

from typing import Any

import numpy as np
import qp

from rail.core.common_params import SHARED_PARAMS
from rail.core.data import (DataHandle, ModelHandle, QPHandle, TableHandle,
                            TableLike)
from rail.core.stage import RailStage

# for backwards compatibility


class CatSummarizer(RailStage):  # pragma: no cover
    """The base class for classes that go from catalog-like tables
    to ensemble NZ estimates.

    CatSummarizer take as "input" a catalog-like table.  I.e., a
    table with fluxes in photometric bands among the set of columns.

    provide as "output" a QPEnsemble, with per-ensemble n(z).
    """

    name = "CatSummarizer"
    config_options = RailStage.config_options.copy()
    config_options.update(chunk_size=SHARED_PARAMS)
    inputs = [("input", TableHandle)]
    outputs = [("output", QPHandle)]

    def summarize(self, input_data: TableLike) -> DataHandle:
        """The main run method for the summarization, should be implemented
        in the specific subclass.

        This will attach the input_data to this `CatSummarizer`
        (for introspection and provenance tracking).

        Then it will call the run() and finalize() methods, which need to
        be implemented by the sub-classes.

        The run() method will need to register the data that it creates to this `CatSummarizer`
        by using `self.add_data('output', output_data)`.

        Finally, this will return a QPHandle providing access to that output data.

        Parameters
        ----------
        input_data
            Either a dictionary of all input data or a `TableHandle` providing access to the same

        Returns
        -------
        DataHandle
            Ensemble with n(z), and any ancilary data
        """
        self.set_data("input", input_data)
        self.run()
        self.finalize()
        return self.get_handle("output")


class PZSummarizer(RailStage):
    """The base class for classes that go from per-galaxy PZ estimates to ensemble NZ estimates

    PZSummarizer take as "input" a `qp.Ensemble` with per-galaxy PDFs, and
    provide as "output" a QPEnsemble, with per-ensemble n(z).
    """

    name = "PZtoNZSummarizer"
    config_options = RailStage.config_options.copy()
    config_options.update(chunk_size=SHARED_PARAMS)
    inputs = [("model", ModelHandle), ("input", QPHandle)]
    outputs = [("output", QPHandle)]

    def summarize(self, input_data: qp.Ensemble) -> qp.Ensemble:
        """The main run method for the summarization, should be implemented
        in the specific subclass.

        This will attach the input_data to this `PZtoNZSummarizer`
        (for introspection and provenance tracking).

        Then it will call the run() and finalize() methods, which need to
        be implemented by the sub-classes.

        The run() method will need to register the data that it creates to this Estimator
        by using `self.add_data('output', output_data)`.

        Finally, this will return a QPHandle providing access to that output data.

        Parameters
        ----------
        input_data
            Per-galaxy p(z), and any ancilary data associated with it

        Returns
        -------
        qp.Ensemble
            Ensemble with n(z), and any ancilary data
        """
        self.set_data("input", input_data)
        self.run()
        self.finalize()
        return self.get_handle("output")

    def _broadcast_bootstrap_matrix(self) -> np.ndarray | None:
        rng = np.random.default_rng(seed=self.config.seed)
        # Only one of the nodes needs to produce the bootstrap indices
        ngal = self._input_length
        if self.rank == 0:
            bootstrap_matrix = rng.integers(
                low=0, high=ngal, size=(ngal, self.config.nsamples)
            )
        else:  # pragma: no cover
            bootstrap_matrix = None
        if self.comm is not None:  # pragma: no cover
            self.comm.Barrier()
            bootstrap_matrix = self.comm.bcast(bootstrap_matrix, root=0)
        return bootstrap_matrix

    def _join_histograms(
        self, bvals: np.ndarray, yvals: np.ndarray
    ) -> tuple[np.ndarray, np.ndarray]:  # pragma: no cover
        bvals_r = self.comm.reduce(bvals)
        yvals_r = self.comm.reduce(yvals)
        return (bvals_r, yvals_r)


class SZPZSummarizer(RailStage):  # pragma: no cover
    """The base class for classes that use two sets of data: a photometry sample with
    spec-z values, and a photometry sample with unknown redshifts, e.g. minisom_som and
    outputs a QP Ensemble with bootstrap realization of the N(z) distribution
    """

    name = "SZPZtoNZSummarizer"
    config_options = RailStage.config_options.copy()
    config_options.update(chunk_size=SHARED_PARAMS)
    inputs = [
        ("input", TableHandle),
        ("spec_input", TableHandle),
        ("model", ModelHandle),
    ]
    outputs = [("output", QPHandle)]

    def __init__(self, args: Any, **kwargs: Any) -> None:
        """Initialize Estimator that can sample galaxy data."""
        super().__init__(args, **kwargs)
        self.model = None
        # NOTE: open model removed from init, need to put an
        # `open_model` call explicitly in the run method for
        # each summarizer.

    def summarize(self, input_data: qp.Ensemble, spec_data: np.ndarray) -> qp.Ensemble:
        """The main run method for the summarization, should be implemented
        in the specific subclass.

        This will attach the input_data to this `SZandPhottoNZSummarizer`
        (for introspection and provenance tracking).

        Then it will call the run() and finalize() methods, which need to
        be implemented by the sub-classes.

        The run() method will need to register the data that it creates to this Estimator
        by using `self.add_data('output', output_data)`.

        Finally, this will return a QPHandle providing access to that output data.

        Parameters
        ----------
        input_data
            Per-galaxy p(z), and any ancilary data associated with it

        spec_data
            Spectroscopic data

        Returns
        -------
        qp.Ensemble
            Ensemble with n(z), and any ancilary data
        """
        self.set_data("input", input_data)
        self.set_data("spec_input", spec_data)
        self.run()
        self.finalize()
        return self.get_handle("output")



import numpy as np
import qp

from rail.evaluation.stats_groups import stat_and_pval

from .base import MetricEvaluator


class CDELoss(MetricEvaluator):
    """Conditional density loss"""

    def __init__(
        self, qp_ens: qp.Ensemble, zgrid: np.ndarray, ztrue: np.ndarray
    ) -> None:
        """Class constructor"""
        super().__init__(qp_ens)

        self._pdfs = qp_ens.pdf(zgrid)
        self._xvals = zgrid
        self._ztrue = ztrue
        self._npdf = qp_ens.npdf

    def evaluate(self) -> stat_and_pval:
        """Evaluate the estimated conditional density loss described in
        Izbicki & Lee 2017 (arXiv:1704.08095).

        Notes
        -----
        """

        # Calculate first term E[\int f*(z | X)^2 dz]
        term1 = np.mean(np.trapz(self._pdfs**2, x=self._xvals))
        # z bin closest to ztrue
        nns = [np.argmin(np.abs(self._xvals - z)) for z in self._ztrue]
        # Calculate second term E[f*(Z | X)]
        term2 = np.mean(self._pdfs[range(self._npdf), nns])
        cdeloss = term1 - 2 * term2
        return stat_and_pval(cdeloss, np.nan)



"""
A summarizer-like stage that simple makes a histogram the true nz
"""

from typing import Any, Generator

import numpy as np
import qp
from ceci.config import StageParameter as Param

from rail.core.common_params import SHARED_PARAMS
from rail.core.data import DataHandle, QPHandle, TableHandle, TableLike
from rail.core.stage import RailStage


class TrueNZHistogrammer(RailStage):
    """Summarizer-like stage which simply histograms the true redshift"""

    name = "TrueNZHistogrammer"
    config_options = RailStage.config_options.copy()
    config_options.update(
        zmin=SHARED_PARAMS,
        zmax=SHARED_PARAMS,
        nzbins=SHARED_PARAMS,
        redshift_col=SHARED_PARAMS,
        selected_bin=Param(int, -1, msg="Which tomography bin to consider"),
        chunk_size=SHARED_PARAMS,
        hdf5_groupname=SHARED_PARAMS,
    )
    inputs = [("input", TableHandle), ("tomography_bins", TableHandle)]
    outputs = [("true_NZ", QPHandle)]

    def __init__(self, args: Any, **kwargs: Any) -> None:
        super().__init__(args, **kwargs)
        self.zgrid: np.ndarray | None = None
        self.bincents: np.ndarray | None = None

    def _setup_iterator(self) -> Generator:
        itrs = [
            self.input_iterator("input", groupname=self.config.hdf5_groupname),
            self.input_iterator("tomography_bins", groupname=""),
        ]

        for it in zip(*itrs):
            first = True
            mask = None
            for s, e, d in it:
                if first:
                    start = s
                    end = e
                    pz_data = d
                    first = False
                else:
                    if self.config.selected_bin < 0:
                        mask = np.ones(e - s, dtype=bool)
                    else:
                        mask = d["class_id"] == self.config.selected_bin
            yield start, end, pz_data, mask

    def run(self) -> None:
        iterator = self._setup_iterator()
        self.zgrid = np.linspace(
            self.config.zmin, self.config.zmax, self.config.nzbins + 1
        )
        assert self.zgrid is not None
        self.bincents = 0.5 * (self.zgrid[1:] + self.zgrid[:-1])
        # Initiallizing the histograms
        single_hist = np.zeros(self.config.nzbins)

        first = True
        for s, e, data, mask in iterator:
            print(f"Process {self.rank} running estimator on chunk {s:,} - {e:,}")
            self._process_chunk(s, e, data, mask, first, single_hist)
            first = False
        if self.comm is not None:  # pragma: no cover
            single_hist = self.comm.reduce(single_hist)

        if self.rank == 0:
            n_total = single_hist.sum()
            qp_d = qp.Ensemble(
                qp.hist,
                data=dict(bins=self.zgrid, pdfs=np.atleast_2d(single_hist)),
                ancil=dict(n_total=np.array([n_total], dtype=int)),
            )
            self.add_data("true_NZ", qp_d)

    def _process_chunk(
        self,
        _start: int,
        _end: int,
        data: TableLike,
        mask: np.ndarray,
        _first: bool,
        single_hist: np.ndarray,
    ) -> None:
        squeeze_mask = np.squeeze(mask)
        zb = data[self.config.redshift_col][squeeze_mask]
        assert self.zgrid is not None
        single_hist += np.histogram(zb, bins=self.zgrid)[0]

    def histogram(self, catalog: TableLike, tomo_bins: TableLike) -> DataHandle:
        """The main interface method for ``TrueNZHistogrammer``.

        Creates histogram of N of Z_true.

        This will attach the sample to this `Stage` (for introspection and
        provenance tracking).

        Then it will call the run() and finalize() methods, which need to be
        implemented by the sub-classes.

        The run() method will need to register the data that it creates to this
        Estimator by using ``self.add_data('output', output_data)``.

        Finally, this will return a PqHandle providing access to that output
        data.

        Parameters
        ----------
        catalog
            The sample with the true NZ column

        tomo_bins
            Tomographic bin assignemnets

        Returns
        -------
        DataHandle
            A handle giving access to a the histogram in QP format
        """
        self.set_data("input", catalog)
        self.set_data("tomography_bins", tomo_bins)
        self.run()
        self.finalize()
        return self.get_handle("true_NZ")
